# until this is merged:
# https://github.com/digma-ai/opentelemetry-instrumentation-digma/pull/41

# stdlib
import asyncio
from functools import wraps
import inspect
from typing import Callable
from typing import ClassVar
from typing import Dict
from typing import Optional
from typing import TypeVar
from typing import Union
from typing import cast

# third party
from opentelemetry import trace
from opentelemetry.semconv.trace import SpanAttributes
from opentelemetry.trace import Tracer
from opentelemetry.trace.span import Span


class TracingDecoratorOptions:
    class NamingSchemes:
        @staticmethod
        def function_qualified_name(func: Callable) -> str:
            return func.__qualname__

        default_scheme = function_qualified_name

    naming_scheme: ClassVar[Callable[[Callable], str]] = NamingSchemes.default_scheme
    default_attributes: ClassVar[Dict[str, str]] = {}

    @classmethod
    def set_naming_scheme(cls, naming_scheme: Callable[[Callable], str]) -> None:
        cls.naming_scheme = naming_scheme

    @classmethod
    def set_default_attributes(
        cls, attributes: Optional[Dict[str, str]] = None
    ) -> None:
        if attributes is not None:
            for att in attributes:
                cls.default_attributes[att] = attributes[att]


T = TypeVar("T", bound=Union[Callable, type])


def instrument(
    _func_or_class: T,
    *,
    span_name: str = "",
    record_exception: bool = True,
    attributes: Optional[Dict[str, str]] = None,
    existing_tracer: Optional[Tracer] = None,
    ignore: bool = False,
) -> T:
    """
    A decorator to instrument a class or function with an OTEL tracing span.
    :param cls: internal, used to specify scope of instrumentation
    :param _func_or_class: The function or span to instrument, this is automatically assigned
    :param span_name: Specify the span name explicitly, rather than use the naming convention.
    This parameter has no effect for class decorators: str
    :param record_exception: Sets whether any exceptions occurring in the span and the stacktrace are recorded
    automatically: bool
    :param attributes:A dictionary of span attributes. These will be automatically added to the span. If defined on a
    class decorator, they will be added to every function span under the class.: dict
    :param existing_tracer: Use a specific tracer instead of creating one :Tracer
    :param ignore: Do not instrument this function, has no effect for class decorators:bool
    :return:The decorator function
    """

    def decorate_class(cls: T) -> T:
        for name, method in inspect.getmembers(cls, inspect.isfunction):
            # Ignore private functions, TODO: maybe make this a setting?
            if not name.startswith("_"):
                if isinstance(inspect.getattr_static(cls, name), staticmethod):
                    setattr(
                        cls,
                        name,
                        staticmethod(
                            instrument(
                                method,
                                record_exception=record_exception,
                                attributes=attributes,
                                existing_tracer=existing_tracer,
                            )
                        ),
                    )
                else:
                    setattr(
                        cls,
                        name,
                        instrument(
                            method,
                            record_exception=record_exception,
                            attributes=attributes,
                            existing_tracer=existing_tracer,
                        ),
                    )

        return cls

    # Check if this is a span or class decorator
    if inspect.isclass(_func_or_class):
        return decorate_class(_func_or_class)

    def span_decorator(func_or_class: T) -> T:
        if inspect.isclass(func_or_class):
            return decorate_class(func_or_class)

        # sig = inspect.signature(func_or_class)

        # Check if already decorated (happens if both class and function
        # decorated). If so, we keep the function decorator settings only
        undecorated_func = getattr(func_or_class, "__tracing_unwrapped__", None)
        if undecorated_func:
            # We have already decorated this function, override
            return func_or_class

        func_or_class.__tracing_unwrapped__ = func_or_class

        tracer = existing_tracer or trace.get_tracer(func_or_class.__module__)

        def _set_semantic_attributes(span: Span, func: Callable) -> None:
            span.set_attribute(SpanAttributes.CODE_NAMESPACE, func.__module__)
            span.set_attribute(SpanAttributes.CODE_FUNCTION, func.__qualname__)
            span.set_attribute(SpanAttributes.CODE_FILEPATH, func.__code__.co_filename)
            span.set_attribute(SpanAttributes.CODE_LINENO, func.__code__.co_firstlineno)

        def _set_attributes(
            span: Span, attributes_dict: Optional[Dict[str, str]] = None
        ) -> None:
            if attributes_dict is not None:
                for att in attributes_dict:
                    span.set_attribute(att, attributes_dict[att])

        @wraps(func_or_class)
        def wrap_with_span_sync(*args, **kwargs):  # type: ignore
            name = span_name or TracingDecoratorOptions.naming_scheme(func_or_class)
            with tracer.start_as_current_span(
                name, record_exception=record_exception
            ) as span:
                _set_semantic_attributes(span, func_or_class)
                _set_attributes(span, TracingDecoratorOptions.default_attributes)
                _set_attributes(span, attributes)
                return func_or_class(*args, **kwargs)

        @wraps(func_or_class)
        async def wrap_with_span_async(*args, **kwargs):  # type: ignore
            name = span_name or TracingDecoratorOptions.naming_scheme(func_or_class)
            with tracer.start_as_current_span(
                name, record_exception=record_exception
            ) as span:
                _set_semantic_attributes(span, func_or_class)
                _set_attributes(span, TracingDecoratorOptions.default_attributes)
                _set_attributes(span, attributes)
                return await func_or_class(*args, **kwargs)

        if ignore:
            return func_or_class

        wrapper = (
            wrap_with_span_async
            if asyncio.iscoroutinefunction(func_or_class)
            else wrap_with_span_sync
        )
        wrapper.__signature__ = inspect.signature(func_or_class)

        return cast(T, wrapper)

    return span_decorator(_func_or_class)



# stdlib
import contextlib
from hashlib import sha256
import io
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

# third party
import docker
from packaging import version
from pydantic import validator
from typing_extensions import Self
import yaml

# relative
from ..serde.serializable import serializable
from ..service.response import SyftError
from ..service.response import SyftSuccess
from ..types.base import SyftBaseModel
from .utils import iterator_to_string

PYTHON_DEFAULT_VER = "3.11"
PYTHON_MIN_VER = version.parse("3.10")
PYTHON_MAX_VER = version.parse("3.12")


def _malformed_python_package_error_msg(pkg: str, name: str = "package_name") -> str:
    return f'You must pin the package to an exact version. Got "{pkg}" expected "{name}==x.y.z"'


class CustomBuildConfig(SyftBaseModel):
    gpu: bool = False
    # python_version: str = PYTHON_DEFAULT_VER
    python_packages: List[str] = []
    system_packages: List[str] = []
    custom_cmds: List[str] = []

    # @validator("python_version")
    # def validate_python_version(cls, ver: str) -> str:
    #     parsed_ver = version.parse(ver)

    #     # TODO: Check if Wolfi OS/apk supports minor version of python
    #     if parsed_ver.micro != 0:
    #         raise ValueError("Provide only major.minor version of python")

    #     if PYTHON_MIN_VER <= parsed_ver < PYTHON_MAX_VER:
    #         return ver
    #     else:
    #         raise ValueError(
    #             f"Python version must be between {PYTHON_MIN_VER} and {PYTHON_MAX_VER}"
    #         )

    @validator("python_packages")
    def validate_python_packages(cls, pkgs: List[str]) -> List[str]:
        for pkg in pkgs:
            ver_parts = ()
            name_ver = pkg.split("==")
            if len(name_ver) != 2:
                raise ValueError(_malformed_python_package_error_msg(pkg))

            name, ver = name_ver

            if ver:
                ver_parts = ver.split(".")

            if not ver or len(ver_parts) <= 2:
                raise ValueError(_malformed_python_package_error_msg(pkg, name))

        return pkgs

    def merged_python_pkgs(self, sep=" ") -> str:
        return sep.join(self.python_packages)

    def merged_system_pkgs(self, sep=" ") -> str:
        return sep.join(self.system_packages)

    def merged_custom_cmds(self, sep=";") -> str:
        return sep.join(self.custom_cmds)


class WorkerConfig(SyftBaseModel):
    pass


@serializable()
class CustomWorkerConfig(WorkerConfig):
    build: CustomBuildConfig
    version: str = "1"

    @classmethod
    def from_dict(cls, config: Dict[str, Any]) -> Self:
        return cls(**config)

    @classmethod
    def from_str(cls, content: str) -> Self:
        config = yaml.safe_load(content)
        return cls.from_dict(config)

    @classmethod
    def from_path(cls, path: Union[Path, str]) -> Self:
        with open(path) as f:
            config = yaml.safe_load(f)
            return cls.from_dict(config)

    def get_signature(self) -> str:
        return sha256(self.json(sort_keys=True).encode()).hexdigest()


@serializable()
class PrebuiltWorkerConfig(WorkerConfig):
    # tag that is already built and pushed in some registry
    tag: str
    description: Optional[str]

    def __str__(self) -> str:
        if self.description:
            return f"prebuilt tag='{self.tag}' description='{self.description}'"
        else:
            return f"prebuilt tag='{self.tag}'"

    def set_description(self, description_text: str) -> None:
        self.description = description_text


@serializable()
class DockerWorkerConfig(WorkerConfig):
    dockerfile: str
    file_name: Optional[str]
    description: Optional[str]

    @validator("dockerfile")
    def validate_dockerfile(cls, dockerfile: str) -> str:
        if not dockerfile:
            raise ValueError("Dockerfile cannot be empty")
        dockerfile = dockerfile.strip()
        return dockerfile

    @classmethod
    def from_path(
        cls,
        path: Union[Path, str],
        description: Optional[str] = "",
    ) -> Self:
        with open(path) as f:
            return cls(
                dockerfile=f.read(),
                file_name=Path(path).name,
                description=description,
            )

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, DockerWorkerConfig):
            return False
        return self.dockerfile == __value.dockerfile

    def __hash__(self) -> int:
        return hash(self.dockerfile)

    def __str__(self) -> str:
        return self.dockerfile

    def set_description(self, description_text: str) -> None:
        self.description = description_text

    def test_image_build(self, tag: str, **kwargs) -> Union[SyftSuccess, SyftError]:
        try:
            with contextlib.closing(docker.from_env()) as client:
                if not client.ping():
                    return SyftError(
                        "Cannot reach docker server. Please check if docker is running."
                    )

                kwargs["fileobj"] = io.BytesIO(self.dockerfile.encode("utf-8"))
                _, logs = client.images.build(
                    tag=tag,
                    **kwargs,
                )
                return SyftSuccess(message=iterator_to_string(iterator=logs))
        except Exception as e:
            return SyftError(message=f"Failed to build: {e}")



# stdlib
from typing import Dict
from typing import Union

# relative
from ...client.enclave_client import EnclaveClient
from ...client.enclave_client import EnclaveMetadata
from ...serde.serializable import serializable
from ...service.response import SyftError
from ...service.response import SyftSuccess
from ...service.user.user_roles import GUEST_ROLE_LEVEL
from ...store.document_store import DocumentStore
from ...types.twin_object import TwinObject
from ...types.uid import UID
from ..action.action_object import ActionObject
from ..code.user_code_service import UserCode
from ..code.user_code_service import UserCodeStatus
from ..context import AuthedServiceContext
from ..context import ChangeContext
from ..network.routes import route_to_connection
from ..service import AbstractService
from ..service import service_method


# TODO 🟣 Created a generic Enclave Service
# Currently it mainly works only for Azure
@serializable()
class EnclaveService(AbstractService):
    store: DocumentStore

    def __init__(self, store: DocumentStore) -> None:
        self.store = store

    @service_method(
        path="enclave.send_user_code_inputs_to_enclave",
        name="send_user_code_inputs_to_enclave",
        roles=GUEST_ROLE_LEVEL,
    )
    def send_user_code_inputs_to_enclave(
        self,
        context: AuthedServiceContext,
        user_code_id: UID,
        inputs: Dict,
        node_name: str,
        node_id: UID,
    ) -> Union[SyftSuccess, SyftError]:
        if not context.node or not context.node.signing_key:
            return SyftError(message=f"{type(context)} has no node")

        root_context = AuthedServiceContext(
            credentials=context.node.verify_key, node=context.node
        )

        user_code_service = context.node.get_service("usercodeservice")
        action_service = context.node.get_service("actionservice")
        user_code = user_code_service.get_by_uid(context=root_context, uid=user_code_id)
        if isinstance(user_code, SyftError):
            return user_code

        reason: str = context.extra_kwargs.get("reason", "")
        status_update = user_code.status.mutate(
            value=(UserCodeStatus.APPROVED, reason),
            node_name=node_name,
            node_id=node_id,
            verify_key=context.credentials,
        )
        if isinstance(status_update, SyftError):
            return status_update

        user_code.status = status_update

        user_code_update = user_code_service.update_code_state(
            context=root_context, code_item=user_code
        )
        if isinstance(user_code_update, SyftError):
            return user_code_update

        root_context = context.as_root_context()
        if not action_service.exists(context=context, obj_id=user_code_id):
            dict_object = ActionObject.from_obj({})
            dict_object.id = user_code_id
            dict_object[str(context.credentials)] = inputs
            root_context.extra_kwargs = {"has_result_read_permission": True}
            # TODO: Instead of using the action store, modify to
            # use the action service directly to store objects
            action_service.set(root_context, dict_object)

        else:
            res = action_service.get(uid=user_code_id, context=root_context)
            if res.is_ok():
                dict_object = res.ok()
                dict_object[str(context.credentials)] = inputs
                action_service.set(root_context, dict_object)
            else:
                return SyftError(
                    message=f"Error while fetching the object on Enclave: {res.err()}"
                )

        return SyftSuccess(message="Enclave Code Status Updated Successfully")


def get_oblv_service():
    # relative
    from ...external import OBLV

    if OBLV:
        # relative
        from ...external.oblv.oblv_service import OblvService

        return OblvService
    else:
        return SyftError(
            message="Oblivious is not enabled."
            "To enable oblivious package, set sy.enable_external_lib('oblv') "
            "on the client side"
            "Or add --oblv when launching by hagrid"
        )


# Checks if the given user code would  propogate value to enclave on acceptance
def propagate_inputs_to_enclave(user_code: UserCode, context: ChangeContext):
    # Temporarily disable Oblivious Enclave
    # from ...external.oblv.deployment_client import OblvMetadata

    # if isinstance(user_code.enclave_metadata, OblvMetadata):
    #     # relative
    #     oblv_service_class = get_oblv_service()
    #     if isinstance(oblv_service_class, SyftError):
    #         return oblv_service_class
    #     method = context.node.get_service_method(oblv_service_class.get_api_for)

    #     api = method(
    #         user_code.enclave_metadata,
    #         context.node.signing_key,
    #         worker_name=context.node.name,
    #     )
    #     send_method = api.services.oblv.send_user_code_inputs_to_enclave

    if isinstance(user_code.enclave_metadata, EnclaveMetadata):
        # TODO 🟣 Restructure url it work for local mode host.docker.internal

        connection = route_to_connection(user_code.enclave_metadata.route)
        enclave_client = EnclaveClient(
            connection=connection,
            credentials=context.node.signing_key,
        )

        send_method = (
            enclave_client.api.services.enclave.send_user_code_inputs_to_enclave
        )

    else:
        return SyftSuccess(message="Current Request does not require Enclave Transfer")

    inputs = user_code.input_policy._inputs_for_context(context)
    if isinstance(inputs, SyftError):
        return inputs

    # Save inputs to blob store
    for var_name, var_value in inputs.items():
        if isinstance(var_value, (ActionObject, TwinObject)):
            # Set the obj location to enclave
            var_value._set_obj_location_(
                enclave_client.api.node_uid,
                enclave_client.verify_key,
            )
            var_value._save_to_blob_storage()

            inputs[var_name] = var_value

    # send data of the current node to enclave
    res = send_method(
        user_code_id=user_code.id,
        inputs=inputs,
        node_name=context.node.name,
        node_id=context.node.id,
    )
    return res



# stdlib
from collections import defaultdict
import datetime
import json
from pathlib import Path
import threading
import time
from typing import Callable
from typing import Dict
from typing import Optional
import uuid

# third party
from pydantic import BaseModel
import redis
from sherlock.lock import BaseLock
from sherlock.lock import FileLock
from sherlock.lock import RedisLock

# relative
from ..serde.serializable import serializable

THREAD_FILE_LOCKS: Dict[int, Dict[str, int]] = defaultdict(dict)


@serializable()
class LockingConfig(BaseModel):
    """
    Locking config

    Args:
        lock_name: str
            Lock name
        namespace: Optional[str]
            Namespace to use for setting lock keys in the backend store.
        expire: Optional[int]
            Lock expiration time in seconds. If explicitly set to `None`, lock will not expire.
        timeout: Optional[int]
             Timeout to acquire lock(seconds)
        retry_interval: float
            Retry interval to retry acquiring a lock if previous attempts failed.
    """

    lock_name: str = "syft_lock"
    namespace: Optional[str] = None
    expire: Optional[int] = 60
    timeout: Optional[int] = 30
    retry_interval: float = 0.1


@serializable()
class NoLockingConfig(LockingConfig):
    """
    No-locking policy
    """

    pass


@serializable()
class ThreadingLockingConfig(LockingConfig):
    """
    Threading-based locking policy
    """

    pass


@serializable()
class FileLockingConfig(LockingConfig):
    """File locking policy"""

    client_path: Optional[Path] = None


@serializable()
class RedisClientConfig(BaseModel):
    host: str = "localhost"
    port: int = 6379
    db: int = 0
    username: Optional[str] = None
    password: Optional[str] = None


@serializable()
class RedisLockingConfig(LockingConfig):
    """Redis locking policy"""

    client: RedisClientConfig = RedisClientConfig()


class ThreadingLock(BaseLock):
    """
    Threading-based Lock. Used to provide the same API as the rest of the locks.
    """

    def __init__(self, expire: int, **kwargs):
        self.expire = expire
        self.locked_timestamp = 0
        self.lock = threading.Lock()

    @property
    def _locked(self):
        """
        Implementation of method to check if lock has been acquired. Must be
        :returns: if the lock is acquired or not
        :rtype: bool
        """
        locked = self.lock.locked()
        if (
            locked
            and time.time() - self.locked_timestamp >= self.expire
            and self.expire != -1
        ):
            self._release()

        return self.lock.locked()

    def _acquire(self):
        """
        Implementation of acquiring a lock in a non-blocking fashion.
        :returns: if the lock was successfully acquired or not
        :rtype: bool
        """
        locked = self.lock.locked()
        if (
            locked
            and time.time() - self.locked_timestamp > self.expire
            and self.expire != -1
        ):
            self._release()

        status = self.lock.acquire(
            blocking=False
        )  # timeout/retries handle in the `acquire` method
        if status:
            self.locked_timestamp = time.time()
        return status

    def _release(self):
        """
        Implementation of releasing an acquired lock.
        """

        try:
            return self.lock.release()
        except RuntimeError:  # already unlocked
            pass

    def _renew(self) -> bool:
        """
        Implementation of renewing an acquired lock.
        """
        return True


class PatchedFileLock(FileLock):
    """
    Implementation of lock with the file system as the backend for synchronization.
    This version patches for the `FileLock._expiry_time` crash(https://github.com/py-sherlock/sherlock/issues/71)

    `sherlock.FileLock` might not work as expected for Python threads.
    It uses re-entrant OS locks, meaning that multiple Python threads could acquire the lock at the same time.
    For different processes/OS threads, the file lock will work as expected.
    We need to patch the lock to handle Python threads too.

    """

    def __init__(self, *args, **kwargs) -> None:
        self._lock_file_enabled = True
        try:
            super().__init__(*args, **kwargs)
        except BaseException as e:
            print(f"Failed to create a file lock = {e}. Using memory-lock only")
            self._lock_file_enabled = False

        self._lock_py_thread = ThreadingLock(*args, **kwargs)

    def _expiry_time(self) -> str:
        if self.expire is not None:
            expiry_time = self._now() + datetime.timedelta(seconds=self.expire)
        else:
            expiry_time = datetime.datetime.max.replace(
                tzinfo=datetime.timezone.utc
            ).astimezone(datetime.timezone.utc)
        return expiry_time.isoformat()

    def _thread_safe_cbk(self, cbk: Callable) -> bool:
        # Acquire lock at Python level(if-needed)
        locked = self._lock_py_thread._acquire()
        if not locked:
            return False

        try:
            result = cbk()
        except BaseException as e:
            print(e)
            result = False

        self._lock_py_thread._release()
        return result

    def _acquire(self) -> bool:
        return self._thread_safe_cbk(self._acquire_file_lock)

    def _release(self) -> None:
        res = self._thread_safe_cbk(self._release_file_lock)
        return res

    def _acquire_file_lock(self) -> bool:
        if not self._lock_file_enabled:
            return True

        owner = str(uuid.uuid4())

        # Acquire lock at OS level
        with self._lock_file:
            if self._data_file.exists():
                for _retry in range(10):
                    try:
                        data = json.loads(self._data_file.read_text())
                        break
                    except BaseException:
                        time.sleep(0.1)
                    if _retry == 9:
                        pass

                now = self._now()
                has_expired = self._has_expired(data, now)
                if owner != data["owner"]:
                    if not has_expired:
                        # Someone else holds the lock.
                        return False
                    else:
                        # Lock is available for us to take.
                        data = {"owner": owner, "expiry_time": self._expiry_time()}
                else:
                    # Same owner so do not set or modify Lease.
                    return False
            else:
                data = {"owner": owner, "expiry_time": self._expiry_time()}

            # Write new data back to file.
            self._data_file.touch()
            self._data_file.write_text(json.dumps(data))

            # We succeeded in writing to the file so we now hold the lock.
            self._owner = owner

            return True

    @property
    def _locked(self):
        if self._lock_py_thread.locked():
            return True

        if not self._lock_file_enabled:
            return False

        if not self._data_file.exists():
            # File doesn't exist so can't be locked.
            return False

        with self._lock_file:
            data = None
            for _retry in range(10):
                try:
                    data = json.loads(self._data_file.read_text())
                    break
                except BaseException:
                    time.sleep(0.1)

        if data is None:
            return False

        if self._has_expired(data, self._now()):
            # File exists but has expired.
            return False

        # Lease exists and has not expired.
        return True

    def _release_file_lock(self) -> None:
        if not self._lock_file_enabled:
            return

        if self._owner is None:
            return

        if not self._data_file.exists():
            return

        with self._lock_file:
            data = None
            for _retry in range(10):
                try:
                    data = json.loads(self._data_file.read_text())
                    break
                except BaseException:
                    time.sleep(0.1)

            if data is None:
                return

            if self._owner == data["owner"]:
                self._data_file.unlink()
                self._owner = None


class SyftLock(BaseLock):
    """
    Syft Lock implementations.

    Params:
        config: Config specific to a locking strategy.
    """

    def __init__(self, config: LockingConfig):
        self.config = config

        self.lock_name = config.lock_name
        self.namespace = config.namespace
        self.expire = config.expire
        self.timeout = config.timeout
        self.retry_interval = config.retry_interval

        self.passthrough = False

        self._lock: Optional[BaseLock] = None

        base_params = {
            "lock_name": config.lock_name,
            "namespace": config.namespace,
            "expire": config.expire,
            "timeout": config.timeout,
            "retry_interval": config.retry_interval,
        }
        if isinstance(config, NoLockingConfig):
            self.passthrough = True
        elif isinstance(config, ThreadingLockingConfig):
            self._lock = ThreadingLock(**base_params)
        elif isinstance(config, FileLockingConfig):
            client = config.client_path
            self._lock = PatchedFileLock(
                **base_params,
                client=client,
            )
        elif isinstance(config, RedisLockingConfig):
            client = redis.StrictRedis(**config.client.dict())

            self._lock = RedisLock(
                **base_params,
                client=client,
            )
        else:
            raise ValueError("Unsupported config type")

    @property
    def _locked(self):
        """
        Implementation of method to check if lock has been acquired.

        :returns: if the lock is acquired or not
        :rtype: bool
        """
        if self.passthrough:
            return False

        return self._lock.locked()

    def acquire(self, blocking: bool = True) -> bool:
        """
        Acquire a lock, blocking or non-blocking.
        :param bool blocking: acquire a lock in a blocking or non-blocking
                              fashion. Defaults to True.
        :returns: if the lock was successfully acquired or not
        :rtype: bool
        """

        if not blocking:
            return self._acquire()

        timeout = self.timeout
        start_time = time.time()
        elapsed = 0
        while timeout >= elapsed:
            if not self._acquire():
                time.sleep(self.retry_interval)
                elapsed = time.time() - start_time
            else:
                return True
        print(
            f"Timeout elapsed after {self.timeout} seconds while trying to acquiring lock."
        )
        # third party
        return False

    def _acquire(self) -> bool:
        """
        Implementation of acquiring a lock in a non-blocking fashion.
        `acquire` makes use of this implementation to provide blocking and non-blocking implementations.

        :returns: if the lock was successfully acquired or not
        :rtype: bool
        """
        if self.passthrough:
            return True

        try:
            return self._lock._acquire()
        except BaseException:
            return False

    def _release(self):
        """
        Implementation of releasing an acquired lock.
        """
        if self.passthrough:
            return

        try:
            return self._lock._release()
        except BaseException:
            pass

    def _renew(self) -> bool:
        """
        Implementation of renewing an acquired lock.
        """
        if self.passthrough:
            return True

        return self._lock._renew()



# relative
from ...serde.serializable import serializable
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SYFT_OBJECT_VERSION_2
from ...types.syft_object import SyftObject


@serializable()
class SyftLog(SyftObject):
    __canonical_name__ = "SyftLog"
    __version__ = SYFT_OBJECT_VERSION_1

    stdout: str = ""

    def append(self, new_str: str) -> None:
        self.stdout += new_str


@serializable()
class SyftLogV2(SyftLog):
    __canonical_name__ = "SyftLog"
    __version__ = SYFT_OBJECT_VERSION_2

    stderr: str = ""

    def append_error(self, new_str: str) -> None:
        self.stderr += new_str

    def restart(self) -> None:
        self.stderr = ""
        self.stdout = ""



# stdlib
from collections.abc import Collection
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union

# relative
from ...serde.serializable import serializable
from ...store.document_store import DocumentStore
from ...types.dicttuple import DictTuple
from ...types.uid import UID
from ...util.telemetry import instrument
from ..action.action_permissions import ActionObjectPermission
from ..action.action_permissions import ActionPermission
from ..context import AuthedServiceContext
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import SERVICE_TO_TYPES
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user_roles import DATA_OWNER_ROLE_LEVEL
from ..user.user_roles import DATA_SCIENTIST_ROLE_LEVEL
from ..user.user_roles import GUEST_ROLE_LEVEL
from ..warnings import CRUDReminder
from ..warnings import HighSideCRUDWarning
from .dataset import Asset
from .dataset import CreateDataset
from .dataset import Dataset
from .dataset import DatasetPageView
from .dataset_stash import DatasetStash


def _paginate_collection(
    collection: Collection,
    page_size: Optional[int] = 0,
    page_index: Optional[int] = 0,
) -> Optional[slice]:
    if page_size is None or page_size <= 0:
        return None

    # If chunk size is defined, then split list into evenly sized chunks
    total = len(collection)
    page_index = 0 if page_index is None else page_index

    if page_size > total or page_index >= total // page_size or page_index < 0:
        return None

    start = page_size * page_index
    stop = min(page_size * (page_index + 1), total)
    return slice(start, stop)


def _paginate_dataset_collection(
    datasets: Sequence[Dataset],
    page_size: Optional[int] = 0,
    page_index: Optional[int] = 0,
) -> Union[DictTuple[str, Dataset], DatasetPageView]:
    slice_ = _paginate_collection(datasets, page_size=page_size, page_index=page_index)
    chunk = datasets[slice_] if slice_ is not None else datasets
    results = DictTuple(chunk, lambda dataset: dataset.name)

    return (
        results
        if slice_ is None
        else DatasetPageView(datasets=results, total=len(datasets))
    )


@instrument
@serializable()
class DatasetService(AbstractService):
    store: DocumentStore
    stash: DatasetStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = DatasetStash(store=store)

    @service_method(
        path="dataset.add",
        name="add",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def add(
        self, context: AuthedServiceContext, dataset: CreateDataset
    ) -> Union[SyftSuccess, SyftError]:
        """Add a Dataset"""
        dataset = dataset.to(Dataset, context=context)
        result = self.stash.set(
            context.credentials,
            dataset,
            add_permissions=[
                ActionObjectPermission(
                    uid=dataset.id, permission=ActionPermission.ALL_READ
                ),
            ],
        )
        if result.is_err():
            return SyftError(message=str(result.err()))
        return SyftSuccess(
            message=f"Dataset uploaded to '{context.node.name}'. "
            f"To see the datasets uploaded by a client on this node, use command `[your_client].datasets`"
        )

    @service_method(
        path="dataset.get_all",
        name="get_all",
        roles=GUEST_ROLE_LEVEL,
        warning=CRUDReminder(),
    )
    def get_all(
        self,
        context: AuthedServiceContext,
        page_size: Optional[int] = 0,
        page_index: Optional[int] = 0,
    ) -> Union[DatasetPageView, DictTuple[str, Dataset], SyftError]:
        """Get a Dataset"""
        result = self.stash.get_all(context.credentials)
        if not result.is_ok():
            return SyftError(message=result.err())

        datasets = result.ok()

        for dataset in datasets:
            dataset.node_uid = context.node.id

        return _paginate_dataset_collection(
            datasets=datasets, page_size=page_size, page_index=page_index
        )

    @service_method(
        path="dataset.search", name="search", roles=DATA_SCIENTIST_ROLE_LEVEL
    )
    def search(
        self,
        context: AuthedServiceContext,
        name: str,
        page_size: Optional[int] = 0,
        page_index: Optional[int] = 0,
    ) -> Union[DatasetPageView, SyftError]:
        """Search a Dataset by name"""
        results = self.get_all(context)

        if isinstance(results, SyftError):
            return results

        filtered_results = [
            dataset for dataset_name, dataset in results.items() if name in dataset_name
        ]

        return _paginate_dataset_collection(
            filtered_results, page_size=page_size, page_index=page_index
        )

    @service_method(path="dataset.get_by_id", name="get_by_id")
    def get_by_id(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[SyftSuccess, SyftError]:
        """Get a Dataset"""
        result = self.stash.get_by_uid(context.credentials, uid=uid)
        if result.is_ok():
            dataset = result.ok()
            dataset.node_uid = context.node.id
            return dataset
        return SyftError(message=result.err())

    @service_method(path="dataset.get_by_action_id", name="get_by_action_id")
    def get_by_action_id(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[List[Dataset], SyftError]:
        """Get Datasets by an Action ID"""
        result = self.stash.search_action_ids(context.credentials, uid=uid)
        if result.is_ok():
            datasets = result.ok()
            for dataset in datasets:
                dataset.node_uid = context.node.id
            return datasets
        return SyftError(message=result.err())

    @service_method(
        path="dataset.get_assets_by_action_id",
        name="get_assets_by_action_id",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_assets_by_action_id(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[List[Asset], SyftError]:
        """Get Assets by an Action ID"""
        datasets = self.get_by_action_id(context=context, uid=uid)
        assets = []
        if isinstance(datasets, list):
            for dataset in datasets:
                for asset in dataset.asset_list:
                    if asset.action_id == uid:
                        assets.append(asset)
            return assets
        elif isinstance(datasets, SyftError):
            return datasets
        return []

    @service_method(
        path="dataset.delete_by_id",
        name="dataset_delete_by_id",
        warning=HighSideCRUDWarning(confirmation=True),
    )
    def delete_dataset(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[SyftSuccess, SyftError]:
        result = self.stash.delete_by_uid(context.credentials, uid)
        if result.is_ok():
            return result.ok()
        else:
            return SyftError(message=result.err())


TYPE_TO_SERVICE[Dataset] = DatasetService
SERVICE_TO_TYPES[DatasetService].update({Dataset})



# future
from __future__ import annotations

# stdlib
import ast
from copy import deepcopy
from enum import Enum
import hashlib
import inspect
from inspect import Parameter
from inspect import Signature
from io import StringIO
import sys
import types
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

# third party
from RestrictedPython import compile_restricted
from result import Ok

# relative
from ...abstract_node import NodeType
from ...client.api import NodeIdentity
from ...node.credentials import SyftVerifyKey
from ...serde.recursive_primitives import recursive_serde_register_type
from ...serde.serializable import serializable
from ...store.document_store import PartitionKey
from ...types.datetime import DateTime
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ...types.transforms import TransformContext
from ...types.transforms import generate_id
from ...types.transforms import transform
from ...types.twin_object import TwinObject
from ...types.uid import UID
from ...util.util import is_interpreter_jupyter
from ..action.action_object import ActionObject
from ..code.code_parse import GlobalsVisitor
from ..code.unparse import unparse
from ..context import AuthedServiceContext
from ..context import ChangeContext
from ..context import NodeServiceContext
from ..dataset.dataset import Asset
from ..response import SyftError
from ..response import SyftSuccess

PolicyUserVerifyKeyPartitionKey = PartitionKey(
    key="user_verify_key", type_=SyftVerifyKey
)
PyCodeObject = Any


def extract_uid(v: Any) -> UID:
    value = v
    if isinstance(v, ActionObject):
        value = v.id
    if isinstance(v, TwinObject):
        value = v.id

    if not isinstance(value, UID):
        raise Exception(f"Input {v} must have a UID not {type(v)}")
    return value


def filter_only_uids(results: Any) -> Dict[str, Any]:
    if not hasattr(results, "__len__"):
        results = [results]

    if isinstance(results, list):
        output_list = []
        for v in results:
            output_list.append(extract_uid(v))
        return output_list
    elif isinstance(results, dict):
        output_dict = {}
        for k, v in results.items():
            output_dict[k] = extract_uid(v)
        return output_dict
    return extract_uid(results)


class Policy(SyftObject):
    # version
    __canonical_name__ = "Policy"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    init_kwargs: Dict[Any, Any] = {}

    def __init__(self, *args, **kwargs) -> None:
        if "init_kwargs" in kwargs:
            init_kwargs = kwargs["init_kwargs"]
            del kwargs["init_kwargs"]
        else:
            init_kwargs = deepcopy(kwargs)
            if "id" in init_kwargs:
                del init_kwargs["id"]
        super().__init__(init_kwargs=init_kwargs, *args, **kwargs)  # noqa: B026

    @classmethod
    @property
    def policy_code(cls) -> str:
        mro = reversed(cls.mro())
        op_code = ""
        for klass in mro:
            if "Policy" in klass.__name__:
                op_code += inspect.getsource(klass)
                op_code += "\n"
        return op_code

    def public_state() -> None:
        raise NotImplementedError

    @property
    def valid(self) -> Union[SyftSuccess, SyftError]:
        return SyftSuccess(message="Policy is valid.")


@serializable()
class UserPolicyStatus(Enum):
    SUBMITTED = "submitted"
    DENIED = "denied"
    APPROVED = "approved"


def partition_by_node(kwargs: Dict[str, Any]) -> Dict[str, UID]:
    # relative
    from ...client.api import APIRegistry
    from ...client.api import NodeIdentity
    from ...types.twin_object import TwinObject
    from ..action.action_object import ActionObject

    # fetches the all the current api's connected
    api_list = APIRegistry.get_all_api()
    output_kwargs = {}
    for k, v in kwargs.items():
        uid = v
        if isinstance(v, ActionObject):
            uid = v.id
        if isinstance(v, TwinObject):
            uid = v.id
        if isinstance(v, Asset):
            uid = v.action_id
        if not isinstance(uid, UID):
            raise Exception(f"Input {k} must have a UID not {type(v)}")

        _obj_exists = False
        for api in api_list:
            if api.services.action.exists(uid):
                node_identity = NodeIdentity.from_api(api)
                if node_identity not in output_kwargs:
                    output_kwargs[node_identity] = {k: uid}
                else:
                    output_kwargs[node_identity].update({k: uid})

                _obj_exists = True
                break

        if not _obj_exists:
            raise Exception(f"Input data {k}:{uid} does not belong to any Domain")

    return output_kwargs


class InputPolicy(Policy):
    __canonical_name__ = "InputPolicy"
    __version__ = SYFT_OBJECT_VERSION_1

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if "init_kwargs" in kwargs:
            init_kwargs = kwargs["init_kwargs"]
            del kwargs["init_kwargs"]
        else:
            # TODO: remove this tech debt, dont remove the id mapping functionality
            init_kwargs = partition_by_node(kwargs)
        super().__init__(*args, init_kwargs=init_kwargs, **kwargs)

    def filter_kwargs(
        self, kwargs: Dict[Any, Any], context: AuthedServiceContext, code_item_id: UID
    ) -> Dict[Any, Any]:
        raise NotImplementedError

    @property
    def inputs(self) -> Dict[NodeIdentity, Any]:
        return self.init_kwargs

    def _inputs_for_context(self, context: ChangeContext):
        user_node_view = NodeIdentity.from_change_context(context)
        inputs = self.inputs[user_node_view]

        root_context = AuthedServiceContext(
            node=context.node, credentials=context.approving_user_credentials
        ).as_root_context()

        action_service = context.node.get_service("actionservice")
        for var_name, uid in inputs.items():
            action_object = action_service.get(uid=uid, context=root_context)
            if action_object.is_err():
                return SyftError(message=action_object.err())
            action_object_value = action_object.ok()
            # resolve syft action data from blob store
            if isinstance(action_object_value, TwinObject):
                action_object_value.private_obj.syft_action_data  # noqa: B018
                action_object_value.mock_obj.syft_action_data  # noqa: B018
            elif isinstance(action_object_value, ActionObject):
                action_object_value.syft_action_data  # noqa: B018
            inputs[var_name] = action_object_value
        return inputs


def retrieve_from_db(
    code_item_id: UID, allowed_inputs: Dict[str, UID], context: AuthedServiceContext
) -> Dict:
    # relative
    from ...service.action.action_object import TwinMode

    action_service = context.node.get_service("actionservice")
    code_inputs = {}

    # When we are retrieving the code from the database, we need to use the node's
    # verify key as the credentials. This is because when we approve the code, we
    # we allow the private data to be used only for this specific code.
    # but we are not modifying the permissions of the private data

    root_context = AuthedServiceContext(
        node=context.node, credentials=context.node.verify_key
    )
    if context.node.node_type == NodeType.DOMAIN:
        for var_name, arg_id in allowed_inputs.items():
            kwarg_value = action_service._get(
                context=root_context,
                uid=arg_id,
                twin_mode=TwinMode.NONE,
                has_permission=True,
            )
            if kwarg_value.is_err():
                return SyftError(message=kwarg_value.err())
            code_inputs[var_name] = kwarg_value.ok()

    elif context.node.node_type == NodeType.ENCLAVE:
        dict_object = action_service.get(context=root_context, uid=code_item_id)
        if dict_object.is_err():
            return SyftError(message=dict_object.err())
        for value in dict_object.ok().syft_action_data.values():
            code_inputs.update(value)

    else:
        raise Exception(
            f"Invalid Node Type for Code Submission:{context.node.node_type}"
        )
    return Ok(code_inputs)


def allowed_ids_only(
    allowed_inputs: Dict[str, UID],
    kwargs: Dict[str, Any],
    context: AuthedServiceContext,
) -> Dict[str, UID]:
    if context.node.node_type == NodeType.DOMAIN:
        node_identity = NodeIdentity(
            node_name=context.node.name,
            node_id=context.node.id,
            verify_key=context.node.signing_key.verify_key,
        )
        allowed_inputs = allowed_inputs.get(node_identity, {})
    elif context.node.node_type == NodeType.ENCLAVE:
        base_dict = {}
        for key in allowed_inputs.values():
            base_dict.update(key)
        allowed_inputs = base_dict
    else:
        raise Exception(
            f"Invalid Node Type for Code Submission:{context.node.node_type}"
        )
    filtered_kwargs = {}
    for key in allowed_inputs.keys():
        if key in kwargs:
            value = kwargs[key]
            uid = value
            if not isinstance(uid, UID):
                uid = getattr(value, "id", None)

            if uid != allowed_inputs[key]:
                raise Exception(
                    f"Input {type(value)} for {key} not in allowed {allowed_inputs}"
                )
            filtered_kwargs[key] = value
    return filtered_kwargs


@serializable()
class ExactMatch(InputPolicy):
    # version
    __canonical_name__ = "ExactMatch"
    __version__ = SYFT_OBJECT_VERSION_1

    def filter_kwargs(
        self, kwargs: Dict[Any, Any], context: AuthedServiceContext, code_item_id: UID
    ) -> Dict[Any, Any]:
        allowed_inputs = allowed_ids_only(
            allowed_inputs=self.inputs, kwargs=kwargs, context=context
        )
        results = retrieve_from_db(
            code_item_id=code_item_id, allowed_inputs=allowed_inputs, context=context
        )
        return results


@serializable()
class OutputHistory(SyftObject):
    # version
    __canonical_name__ = "OutputHistory"
    __version__ = SYFT_OBJECT_VERSION_1

    output_time: DateTime
    outputs: Optional[Union[List[UID], Dict[str, UID]]]
    executing_user_verify_key: SyftVerifyKey


class OutputPolicy(Policy):
    # version
    __canonical_name__ = "OutputPolicy"
    __version__ = SYFT_OBJECT_VERSION_1

    output_history: List[OutputHistory] = []
    output_kwargs: List[str] = []
    node_uid: Optional[UID]
    output_readers: List[SyftVerifyKey] = []

    def apply_output(
        self,
        context: NodeServiceContext,
        outputs: Any,
    ) -> Any:
        output_uids = filter_only_uids(outputs)
        if isinstance(output_uids, UID):
            output_uids = [output_uids]
        history = OutputHistory(
            output_time=DateTime.now(),
            outputs=output_uids,
            executing_user_verify_key=context.credentials,
        )
        self.output_history.append(history)
        return outputs

    @property
    def outputs(self) -> List[str]:
        return self.output_kwargs

    @property
    def last_output_ids(self) -> List[str]:
        return self.output_history[-1].outputs


@serializable()
class OutputPolicyExecuteCount(OutputPolicy):
    __canonical_name__ = "OutputPolicyExecuteCount"
    __version__ = SYFT_OBJECT_VERSION_1

    count: int = 0
    limit: int

    def apply_output(
        self,
        context: NodeServiceContext,
        outputs: Any,
    ) -> Optional[Any]:
        if self.count < self.limit:
            super().apply_output(context, outputs)
            self.count += 1
            return outputs
        return None

    @property
    def valid(self) -> Union[SyftSuccess, SyftError]:
        is_valid = self.count < self.limit
        if is_valid:
            return SyftSuccess(
                message=f"Policy is still valid. count: {self.count} < limit: {self.limit}"
            )
        return SyftError(
            message=f"Policy is no longer valid. count: {self.count} >= limit: {self.limit}"
        )

    def public_state(self) -> None:
        return {"limit": self.limit, "count": self.count}


@serializable()
class OutputPolicyExecuteOnce(OutputPolicyExecuteCount):
    __canonical_name__ = "OutputPolicyExecuteOnce"
    __version__ = SYFT_OBJECT_VERSION_1

    limit: int = 1


SingleExecutionExactOutput = OutputPolicyExecuteOnce


@serializable()
class CustomPolicy(type):
    # capture the init_kwargs transparently
    def __call__(cls, *args: Any, **kwargs: Any) -> None:
        obj = super().__call__(*args, **kwargs)
        obj.init_kwargs = kwargs
        return obj


recursive_serde_register_type(CustomPolicy)


@serializable()
class CustomOutputPolicy(metaclass=CustomPolicy):
    def apply_output(
        self,
        context: NodeServiceContext,
        outputs: Any,
    ) -> Optional[Any]:
        return outputs


class UserOutputPolicy(OutputPolicy):
    __canonical_name__ = "UserOutputPolicy"
    pass


class UserInputPolicy(InputPolicy):
    __canonical_name__ = "UserInputPolicy"
    pass


class EmpyInputPolicy(InputPolicy):
    __canonical_name__ = "EmptyInputPolicy"
    pass


class CustomInputPolicy(metaclass=CustomPolicy):
    pass


@serializable()
class UserPolicy(Policy):
    __canonical_name__ = "UserPolicy"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    node_uid: Optional[UID]
    user_verify_key: SyftVerifyKey
    raw_code: str
    parsed_code: str
    signature: inspect.Signature
    class_name: str
    unique_name: str
    code_hash: str
    byte_code: PyCodeObject
    status: UserPolicyStatus = UserPolicyStatus.SUBMITTED

    @property
    def byte_code(self) -> Optional[PyCodeObject]:
        return compile_byte_code(self.parsed_code)

    @property
    def policy_code(self) -> str:
        return self.raw_code

    def apply_output(
        self,
        context: NodeServiceContext,
        outputs: Any,
    ) -> Optional[Any]:
        return outputs


def new_getfile(object):
    if not inspect.isclass(object):
        return inspect.getfile(object)

    # Lookup by parent module (as in current inspect)
    if hasattr(object, "__module__"):
        object_ = sys.modules.get(object.__module__)
        if hasattr(object_, "__file__"):
            return object_.__file__

    # If parent module is __main__, lookup by methods (NEW)
    for _, member in inspect.getmembers(object):
        if (
            inspect.isfunction(member)
            and object.__qualname__ + "." + member.__name__ == member.__qualname__
        ):
            return inspect.getfile(member)
    else:
        raise TypeError(f"Source for {object!r} not found")


def get_code_from_class(policy):
    klasses = [inspect.getmro(policy)[0]]  #
    whole_str = ""
    for klass in klasses:
        if is_interpreter_jupyter():
            # third party
            from IPython.core.magics.code import extract_symbols

            cell_code = "".join(inspect.linecache.getlines(new_getfile(klass)))
            class_code = extract_symbols(cell_code, klass.__name__)[0][0]
        else:
            class_code = inspect.getsource(klass)
        whole_str += class_code
    return whole_str


@serializable()
class SubmitUserPolicy(Policy):
    __canonical_name__ = "SubmitUserPolicy"
    __version__ = SYFT_OBJECT_VERSION_1

    id: Optional[UID]
    code: str
    class_name: str
    input_kwargs: List[str]

    def compile(self) -> PyCodeObject:
        return compile_restricted(self.code, "<string>", "exec")

    @staticmethod
    def from_obj(policy_obj: CustomPolicy) -> SubmitUserPolicy:
        user_class = policy_obj.__class__
        init_f_code = user_class.__init__.__code__
        return SubmitUserPolicy(
            code=get_code_from_class(user_class),
            class_name=user_class.__name__,
            input_kwargs=init_f_code.co_varnames[1 : init_f_code.co_argcount],
        )


def hash_code(context: TransformContext) -> TransformContext:
    code = context.output["code"]
    del context.output["code"]
    context.output["raw_code"] = code
    code_hash = hashlib.sha256(code.encode("utf8")).hexdigest()
    context.output["code_hash"] = code_hash
    return context


def generate_unique_class_name(context: TransformContext) -> TransformContext:
    # TODO: Do we need to check if the initial name contains underscores?
    code_hash = context.output["code_hash"]
    service_class_name = context.output["class_name"]
    unique_name = f"{service_class_name}_{context.credentials}_{code_hash}"
    context.output["unique_name"] = unique_name
    return context


def compile_byte_code(parsed_code: str) -> Optional[PyCodeObject]:
    try:
        return compile(parsed_code, "<string>", "exec")
    except Exception as e:
        print("WARNING: to compile byte code", e)
    return None


def process_class_code(raw_code: str, class_name: str) -> str:
    tree = ast.parse(raw_code)
    v = GlobalsVisitor()
    v.visit(tree)
    if len(tree.body) != 1 or not isinstance(tree.body[0], ast.ClassDef):
        raise Exception(
            "Class code should only contain the Class definition for your policy"
        )
    old_class = tree.body[0]
    if len(old_class.bases) != 1 or old_class.bases[0].attr not in [
        CustomInputPolicy.__name__,
        CustomOutputPolicy.__name__,
    ]:
        raise Exception(
            f"Class code should either implement {CustomInputPolicy.__name__} "
            f"or {CustomOutputPolicy.__name__}"
        )

    # TODO: changes the bases
    old_class.bases[0].attr = old_class.bases[0].attr.replace("Custom", "User")

    serializable_name = ast.Name(id="sy.serializable", ctx=ast.Load())
    serializable_decorator = ast.Call(
        func=serializable_name,
        args=[],
        keywords=[],
    )

    new_class = tree.body[0]
    # TODO add this manually
    for stmt in new_class.body:
        if isinstance(stmt, ast.FunctionDef) and stmt.name == "__init__":
            stmt.name = "__user_init__"

    # change the module that the code will reference
    # this is required for the @serializable to mount it in the right path for serde
    new_line = ast.parse('__module__ = "syft.user"')
    new_class.body.append(new_line.body[0])
    new_line = ast.parse(f'__canonical_name__ = "{class_name}"')
    new_class.body.append(new_line.body[0])
    new_line = ast.parse("__version__ = 1")
    new_class.body.append(new_line.body[0])
    new_class.name = class_name
    new_class.decorator_list = [serializable_decorator]
    new_body = []
    new_body.append(
        ast.ImportFrom(
            module="__future__",
            names=[ast.alias(name="annotations", asname="annotations")],
            level=0,
        )
    )
    new_body.append(ast.Import(names=[ast.alias(name="syft", asname="sy")], level=0))
    typing_types = [
        "Any",
        "Callable",
        "ClassVar",
        "Dict",
        "List",
        "Optional",
        "Set",
        "Tuple",
        "Type",
    ]
    for typing_type in typing_types:
        new_body.append(
            ast.ImportFrom(
                module="typing",
                names=[ast.alias(name=typing_type, asname=typing_type)],
                level=0,
            )
        )
    new_body.append(new_class)
    module = ast.Module(new_body, type_ignores=[])
    try:
        return unparse(module)
    except Exception as e:
        print("failed to unparse", e)
        raise e


def check_class_code(context: TransformContext) -> TransformContext:
    # TODO: define the proper checking for this case based on the ideas from UserCode
    # check for no globals
    # check for Policy template -> __init__, apply_output, public_state
    # parse init signature
    # check dangerous libraries, maybe compile_restricted already does that
    try:
        processed_code = process_class_code(
            raw_code=context.output["raw_code"],
            class_name=context.output["unique_name"],
        )
        context.output["parsed_code"] = processed_code
    except Exception as e:
        raise e
    return context


def compile_code(context: TransformContext) -> TransformContext:
    byte_code = compile_byte_code(context.output["parsed_code"])
    if byte_code is None:
        raise Exception(
            "Unable to compile byte code from parsed code. "
            + context.output["parsed_code"]
        )
    return context


def add_credentials_for_key(key: str) -> Callable:
    def add_credentials(context: TransformContext) -> TransformContext:
        context.output[key] = context.credentials
        return context

    return add_credentials


def generate_signature(context: TransformContext) -> TransformContext:
    params = [
        Parameter(name=k, kind=Parameter.POSITIONAL_OR_KEYWORD)
        for k in context.output["input_kwargs"]
    ]
    sig = Signature(parameters=params)
    context.output["signature"] = sig
    return context


@transform(SubmitUserPolicy, UserPolicy)
def submit_policy_code_to_user_code() -> List[Callable]:
    return [
        generate_id,
        hash_code,
        generate_unique_class_name,
        generate_signature,
        check_class_code,
        # compile_code, # don't compile until approved
        add_credentials_for_key("user_verify_key"),
    ]


def add_class_to_user_module(klass: type, unique_name: str) -> type:
    klass.__module__ = "syft.user"
    klass.__name__ = unique_name
    # syft absolute
    import syft as sy

    if not hasattr(sy, "user"):
        user_module = types.ModuleType("user")
        sys.modules["syft"].user = user_module
    user_module = sy.user
    setattr(user_module, unique_name, klass)
    sys.modules["syft"].user = user_module
    return klass


def execute_policy_code(user_policy: UserPolicy):
    stdout_ = sys.stdout
    stderr_ = sys.stderr

    try:
        stdout = StringIO()
        stderr = StringIO()

        sys.stdout = stdout
        sys.stderr = stderr

        class_name = f"{user_policy.unique_name}"
        if class_name in user_policy.__object_version_registry__.keys():
            policy_class = user_policy.__object_version_registry__[class_name]
        else:
            exec(user_policy.byte_code)  # nosec
            policy_class = eval(user_policy.unique_name)  # nosec

        policy_class = add_class_to_user_module(policy_class, user_policy.unique_name)

        sys.stdout = stdout_
        sys.stderr = stderr_

        return policy_class

    except Exception as e:
        print("execute_byte_code failed", e, file=stderr_)

    finally:
        sys.stdout = stdout_
        sys.stderr = stderr_


def load_policy_code(user_policy: UserPolicy) -> Any:
    try:
        policy_class = execute_policy_code(user_policy)
        return policy_class
    except Exception as e:
        raise Exception(f"Exception loading code. {user_policy}. {e}")


def init_policy(user_policy: UserPolicy, init_args: Dict[str, Any]):
    policy_class = load_policy_code(user_policy)
    policy_object = policy_class()
    policy_object.__user_init__(**init_args)
    return policy_object



# stdlib
import base64
from enum import Enum
from functools import cache
import json
import os
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
import kr8s
from kr8s.objects import APIObject
from kr8s.objects import Pod
from kr8s.objects import Secret
from pydantic import BaseModel

# Time after which Job will be deleted
JOB_COMPLETION_TTL = 60

# Kubernetes namespace
KUBERNETES_NAMESPACE = os.getenv("K8S_NAMESPACE", "syft")

# Kubernetes runtime flag
IN_KUBERNETES = os.getenv("CONTAINER_HOST") == "k8s"

# Internal registry URL
DEFAULT_INTERNAL_REGISTRY = f"registry.{KUBERNETES_NAMESPACE}.svc.cluster.local"
INTERNAL_REGISTRY_HOST = os.getenv("INTERNAL_REGISTRY_HOST", DEFAULT_INTERNAL_REGISTRY)


class PodPhase(Enum):
    Pending = "Pending"
    Running = "Running"
    Succeeded = "Succeeded"
    Failed = "Failed"
    Unknown = "Unknown"


class PodCondition(BaseModel):
    pod_scheduled: bool
    containers_ready: bool
    initialized: bool
    ready: bool

    @classmethod
    def from_conditions(cls, conditions: list):
        pod_cond = KubeUtils.list_dict_unpack(conditions, key="type", value="status")
        pod_cond_flags = {k: v == "True" for k, v in pod_cond.items()}
        return cls(
            pod_scheduled=pod_cond_flags.get("PodScheduled", False),
            containers_ready=pod_cond_flags.get("ContainersReady", False),
            initialized=pod_cond_flags.get("Initialized", False),
            ready=pod_cond_flags.get("Ready", False),
        )


class ContainerStatus(BaseModel):
    ready: bool
    running: bool
    waiting: bool
    reason: Optional[str]  # when waiting=True
    message: Optional[str]  # when waiting=True
    startedAt: Optional[str]  # when running=True

    @classmethod
    def from_status(cls, cstatus: dict):
        cstate = cstatus.get("state", {})

        return cls(
            ready=cstatus.get("ready", False),
            running="running" in cstate,
            waiting="waiting" in cstate,
            reason=cstate.get("waiting", {}).get("reason", None),
            message=cstate.get("waiting", {}).get("message", None),
            startedAt=cstate.get("running", {}).get("startedAt", None),
        )


class PodStatus(BaseModel):
    phase: PodPhase
    condition: PodCondition
    container: ContainerStatus

    @classmethod
    def from_status_dict(cls: "PodStatus", status: dict):
        return cls(
            phase=PodPhase(status.get("phase", "Unknown")),
            condition=PodCondition.from_conditions(status.get("conditions", [])),
            container=ContainerStatus.from_status(
                status.get("containerStatuses", {})[0]
            ),
        )


@cache
def get_kr8s_client() -> kr8s.Api:
    if not IN_KUBERNETES:
        raise RuntimeError("Not inside a kubernetes environment")
    return kr8s.api(namespace=KUBERNETES_NAMESPACE)


class KubeUtils:
    """
    This class contains utility functions for interacting with kubernetes objects.

    DO NOT call `get_kr8s_client()` inside this class, instead pass it as an argument to the functions.
    This is to avoid calling these functions on resources across namespaces!
    """

    @staticmethod
    def resolve_pod(client: kr8s.Api, pod: Union[str, Pod]) -> Optional[Pod]:
        """Return the first pod that matches the given name"""
        if isinstance(pod, Pod):
            return pod

        for _pod in client.get("pods", pod):
            return _pod

    @staticmethod
    def get_logs(pods: List[Pod]):
        """Combine and return logs for all the pods as string"""
        logs = []
        for pod in pods:
            logs.append(f"----------Logs for pod={pod.metadata.name}----------")
            for log in pod.logs():
                logs.append(log)

        return "\n".join(logs)

    @staticmethod
    def get_pod_status(pod: Pod) -> Optional[PodStatus]:
        """Map the status of the given pod to PodStatuss."""
        if not pod:
            return None
        return PodStatus.from_status_dict(pod.status)

    @staticmethod
    def get_pod_env(pod: Pod) -> Optional[List[Dict]]:
        """Return the environment variables of the first container in the pod."""
        if not pod:
            return

        for container in pod.spec.containers:
            return container.env.to_list()

    @staticmethod
    def get_container_exit_code(pods: List[Pod]) -> List[int]:
        """Return the exit codes of all the containers in the given pods."""
        exit_codes = []
        for pod in pods:
            for container_status in pod.status.containerStatuses:
                exit_codes.append(container_status.state.terminated.exitCode)
        return exit_codes

    @staticmethod
    def get_container_exit_message(pods: List[Pod]) -> Optional[str]:
        """Return the exit message of the first container that exited with non-zero code."""
        for pod in pods:
            for container_status in pod.status.containerStatuses:
                if container_status.state.terminated.exitCode != 0:
                    continue
                return container_status.state.terminated.message
        return None

    @staticmethod
    def b64encode_secret(data: str) -> str:
        """Convert the data to base64 encoded string for Secret."""
        return base64.b64encode(data.encode()).decode()

    @staticmethod
    def create_dockerconfig_secret(
        secret_name: str,
        component: str,
        registries: Iterable[Tuple[str, str, str]],
    ) -> Secret:
        auths = {}

        for url, uname, passwd in registries:
            auths[url] = {
                "username": uname,
                "password": passwd,
                "auth": KubeUtils.b64encode_secret(f"{uname}:{passwd}"),
            }

        config_str = json.dumps({"auths": auths})

        return KubeUtils.create_secret(
            secret_name=secret_name,
            type="kubernetes.io/dockerconfigjson",
            component=component,
            data={
                ".dockerconfigjson": KubeUtils.b64encode_secret(config_str),
            },
        )

    @staticmethod
    def create_secret(
        secret_name: str,
        type: str,
        component: str,
        data: str,
        encoded=True,
    ) -> Secret:
        if not encoded:
            for k, v in data.items():
                data[k] = KubeUtils.b64encode_secret(v)

        secret = Secret(
            {
                "metadata": {
                    "name": secret_name,
                    "labels": {
                        "app.kubernetes.io/name": KUBERNETES_NAMESPACE,
                        "app.kubernetes.io/component": component,
                        "app.kubernetes.io/managed-by": "kr8s",
                    },
                },
                "type": type,
                "data": data,
            }
        )
        return KubeUtils.create_or_get(secret)

    @staticmethod
    def create_or_get(obj: APIObject) -> APIObject:
        if obj.exists():
            obj.refresh()
        else:
            obj.create()
        return obj

    @staticmethod
    def patch_env_vars(env_list: List[Dict], env_dict: Dict) -> List[Dict]:
        """Patch kubernetes pod environment variables in the list with the provided dictionary."""

        # update existing
        for item in env_list:
            k = item["name"]
            if k in env_dict:
                v = env_dict.pop(k)
                item["value"] = v

        # append remaining
        for k, v in env_dict.items():
            env_list.append({"name": k, "value": v})

        return env_list

    @staticmethod
    def list_dict_unpack(
        input_list: List[Dict],
        key: str = "key",
        value: str = "value",
    ) -> Dict:
        # Snapshot from kr8s._data_utils
        return {i[key]: i[value] for i in input_list}



# stdlib
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
from result import Err
from result import Ok

# relative
from ...serde.serializable import serializable
from ...store.document_store import DocumentStore
from ...store.linked_obj import LinkedObject
from ...types.uid import UID
from ...util.telemetry import instrument
from ..action.action_permissions import ActionObjectPermission
from ..action.action_permissions import ActionPermission
from ..code.user_code import UserCode
from ..context import AuthedServiceContext
from ..notification.notification_service import CreateNotification
from ..notification.notification_service import NotificationService
from ..notification.notifications import Notification
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import SERVICE_TO_TYPES
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user import UserView
from ..user.user_roles import GUEST_ROLE_LEVEL
from ..user.user_service import UserService
from .request import Change
from .request import Request
from .request import RequestInfo
from .request import RequestInfoFilter
from .request import RequestStatus
from .request import SubmitRequest
from .request import UserCodeStatusChange
from .request_stash import RequestStash


@instrument
@serializable()
class RequestService(AbstractService):
    store: DocumentStore
    stash: RequestStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = RequestStash(store=store)

    @service_method(path="request.submit", name="submit", roles=GUEST_ROLE_LEVEL)
    def submit(
        self,
        context: AuthedServiceContext,
        request: SubmitRequest,
        send_message: bool = True,
        reason: Optional[str] = "",
    ) -> Union[Request, SyftError]:
        """Submit a Request"""
        try:
            req = request.to(Request, context=context)
            result = self.stash.set(
                context.credentials,
                req,
                add_permissions=[
                    ActionObjectPermission(
                        uid=req.id, permission=ActionPermission.ALL_READ
                    ),
                ],
            )
            if result.is_ok():
                request = result.ok()
                link = LinkedObject.with_context(request, context=context)
                admin_verify_key = context.node.get_service_method(
                    UserService.admin_verify_key
                )

                root_verify_key = admin_verify_key()
                if send_message:
                    subject_msg = f"Result to request {str(request.id)[:4]}...{str(request.id)[-3:]}\
                        has been successfully deposited."
                    message = CreateNotification(
                        subject=subject_msg if not reason else reason,
                        from_user_verify_key=context.credentials,
                        to_user_verify_key=root_verify_key,
                        linked_obj=link,
                    )
                    method = context.node.get_service_method(NotificationService.send)
                    result = method(context=context, notification=message)
                    if isinstance(result, Notification):
                        return Ok(request)
                    else:
                        return Err(result)

                return Ok(request)

            if result.is_err():
                return SyftError(message=str(result.err()))
            return result.ok()
        except Exception as e:
            print("Failed to submit Request", e)
            raise e

    def expand_node(self, context: AuthedServiceContext, code_obj: UserCode):
        user_code_service = context.node.get_service("usercodeservice")
        nested_requests = user_code_service.solve_nested_requests(context, code_obj)

        new_nested_requests = {}
        for func_name, code in nested_requests.items():
            nested_dict = self.expand_node(context, code)
            if isinstance(nested_dict, SyftError):
                return nested_dict
            code.nested_codes = nested_dict
            res = user_code_service.stash.update(context.credentials, code)
            if isinstance(res, Err):
                return res
            linked_obj = LinkedObject.from_obj(code, node_uid=context.node.id)
            new_nested_requests[func_name] = (linked_obj, nested_dict)

        return new_nested_requests

    def resolve_nested_requests(self, context, request):
        # TODO: change this if we have more UserCode Changes
        if len(request.changes) != 1:
            return request

        change = request.changes[0]
        if isinstance(change, UserCodeStatusChange):
            if change.nested_solved:
                return request
            code_obj = change.linked_obj.resolve_with_context(context=context).ok()
            # recursively check what other UserCodes to approve
            nested_requests: Dict[str : Tuple[LinkedObject, Dict]] = self.expand_node(
                context, code_obj
            )
            if isinstance(nested_requests, Err):
                return SyftError(message=nested_requests.value)
            change.nested_solved = True
            code_obj.nested_codes = nested_requests
            change.linked_obj.update_with_context(context=context, obj=code_obj)

            request.changes = [change]
            new_request = self.save(context=context, request=request)
            return new_request
        return request

    @service_method(path="request.get_all", name="get_all")
    def get_all(self, context: AuthedServiceContext) -> Union[List[Request], SyftError]:
        result = self.stash.get_all(context.credentials)
        if result.is_err():
            return SyftError(message=str(result.err()))
        requests = result.ok()
        return [self.resolve_nested_requests(context, request) for request in requests]

    @service_method(path="request.get_all_info", name="get_all_info")
    def get_all_info(
        self,
        context: AuthedServiceContext,
        page_index: Optional[int] = 0,
        page_size: Optional[int] = 0,
    ) -> Union[List[RequestInfo], SyftError]:
        """Get a Dataset"""
        result = self.stash.get_all(context.credentials)
        method = context.node.get_service_method(UserService.get_by_verify_key)
        get_message = context.node.get_service_method(NotificationService.filter_by_obj)

        requests = []
        if result.is_ok():
            for req in result.ok():
                user = method(req.requesting_user_verify_key).to(UserView)
                message = get_message(context=context, obj_uid=req.id)
                requests.append(
                    RequestInfo(user=user, request=req, notification=message)
                )

            # If chunk size is defined, then split list into evenly sized chunks
            if page_size:
                requests = [
                    requests[i : i + page_size]
                    for i in range(0, len(requests), page_size)
                ]
                # Return the proper slice using chunk_index
                requests = requests[page_index]

            return requests

        return SyftError(message=result.err())

    @service_method(path="request.add_changes", name="add_changes")
    def add_changes(
        self, context: AuthedServiceContext, uid: UID, changes: List[Change]
    ) -> Union[Request, SyftError]:
        result = self.stash.get_by_uid(credentials=context.credentials, uid=uid)

        if result.is_err():
            return SyftError(
                message=f"Failed to retrieve request with uid: {uid}. Error: {result.err()}"
            )

        request = result.ok()
        request.changes.extend(changes)
        return self.save(context=context, request=request)

    @service_method(path="request.filter_all_info", name="filter_all_info")
    def filter_all_info(
        self,
        context: AuthedServiceContext,
        request_filter: RequestInfoFilter,
        page_index: Optional[int] = 0,
        page_size: Optional[int] = 0,
    ) -> Union[List[RequestInfo], SyftError]:
        """Get a Dataset"""
        result = self.get_all_info(context)
        requests = list(
            filter(lambda res: (request_filter.name in res.user.name), result)
        )

        # If chunk size is defined, then split list into evenly sized chunks
        if page_size:
            requests = [
                requests[i : i + page_size] for i in range(0, len(requests), page_size)
            ]
            # Return the proper slice using chunk_index
            requests = requests[page_index]

        return requests

    @service_method(
        path="request.apply",
        name="apply",
    )
    def apply(
        self,
        context: AuthedServiceContext,
        uid: UID,
        **kwargs: dict,
    ) -> Union[SyftSuccess, SyftError]:
        request = self.stash.get_by_uid(context.credentials, uid)
        if request.is_ok():
            request = request.ok()

            context.extra_kwargs = kwargs
            result = request.apply(context=context)

            filter_by_obj = context.node.get_service_method(
                NotificationService.filter_by_obj
            )
            request_notification = filter_by_obj(context=context, obj_uid=uid)

            link = LinkedObject.with_context(request, context=context)
            if not request.status == RequestStatus.PENDING:
                mark_as_read = context.node.get_service_method(
                    NotificationService.mark_as_read
                )
                mark_as_read(context=context, uid=request_notification.id)

                notification = CreateNotification(
                    subject=f"{request.changes} for Request id: {uid} has status updated to {request.status}",
                    to_user_verify_key=request.requesting_user_verify_key,
                    linked_obj=link,
                )
                send_notification = context.node.get_service_method(
                    NotificationService.send
                )
                send_notification(context=context, notification=notification)

            # TODO: check whereever we're return SyftError encapsulate it in Result.
            if hasattr(result, "value"):
                return result.value
            return result
        return request.value

    @service_method(path="request.undo", name="undo")
    def undo(
        self, context: AuthedServiceContext, uid: UID, reason: str
    ) -> Union[SyftSuccess, SyftError]:
        result = self.stash.get_by_uid(credentials=context.credentials, uid=uid)
        if result.is_err():
            return SyftError(
                message=f"Failed to update request: {uid} with error: {result.err()}"
            )

        request = result.ok()
        if request is None:
            return SyftError(message=f"Request with uid: {uid} does not exists.")

        context.extra_kwargs["reason"] = reason
        result = request.undo(context=context)

        if result.is_err():
            return SyftError(
                message=f"Failed to undo Request: <{uid}> with error: {result.err()}"
            )

        link = LinkedObject.with_context(request, context=context)
        message_subject = (
            f"Your request for uid: {uid} has been denied. "
            f"Reason specified by Data Owner: {reason}."
        )

        notification = CreateNotification(
            subject=message_subject,
            to_user_verify_key=request.requesting_user_verify_key,
            linked_obj=link,
        )
        send_notification = context.node.get_service_method(NotificationService.send)
        send_notification(context=context, notification=notification)

        return SyftSuccess(message=f"Request {uid} successfully denied !")

    def save(
        self, context: AuthedServiceContext, request: Request
    ) -> Union[Request, SyftError]:
        result = self.stash.update(context.credentials, request)
        if result.is_ok():
            return result.ok()
        return SyftError(
            message=f"Failed to update Request: <{request.id}>. Error: {result.err()}"
        )


TYPE_TO_SERVICE[Request] = RequestService
SERVICE_TO_TYPES[RequestService].update({Request})



# stdlib
import json
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

# relative
from ...client.api import APIRegistry
from ...client.enclave_client import EnclaveMetadata
from ...serde.serializable import serializable
from ...service.user.user_roles import ServiceRole
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ...types.syft_object import SyftVerifyKey
from ...types.syft_object import get_repr_values_table
from ...types.uid import UID
from ...util.notebook_ui.notebook_addons import create_table_template
from ..code.user_code import UserCode
from ..response import SyftError


@serializable()
class CodeHistory(SyftObject):
    # version
    __canonical_name__ = "CodeHistory"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    node_uid: UID
    user_verify_key: SyftVerifyKey
    enclave_metadata: Optional[EnclaveMetadata] = None
    user_code_history: List[UID] = []
    service_func_name: str
    comment_history: List[str] = []

    __attr_searchable__ = ["user_verify_key", "service_func_name"]

    def add_code(self, code: UserCode, comment: Optional[str] = None):
        self.user_code_history.append(code.id)
        if comment is None:
            comment = ""
        self.comment_history.append(comment)


@serializable()
class CodeHistoryView(SyftObject):
    # version
    __canonical_name__ = "CodeHistoryView"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    user_code_history: List[UserCode] = []
    service_func_name: str
    comment_history: List[str] = []

    def _coll_repr_(self):
        return {"Number of versions": len(self.user_code_history)}

    def _repr_html_(self):
        rows = get_repr_values_table(self.user_code_history, True)
        for i, r in enumerate(rows):
            r["Version"] = f"v{i}"
            raw_code = self.user_code_history[i].raw_code
            n_code_lines = raw_code.count("\n")
            if n_code_lines > 5:
                raw_code = "\n".join(raw_code.split("\n", 5))
            r["Code"] = raw_code
        # rows = sorted(rows, key=lambda x: x["Version"])
        return create_table_template(rows, "CodeHistory", table_icon=None)

    def __getitem__(self, index: int):
        api = APIRegistry.api_for(self.syft_node_location, self.syft_client_verify_key)
        if api.user_role.value >= ServiceRole.DATA_OWNER.value:
            if index < 0:
                return SyftError(
                    message="For security concerns we do not allow negative indexing. \
                    Try using absolute values when indexing"
                )
        return self.user_code_history[index]


@serializable()
class CodeHistoriesDict(SyftObject):
    # version
    __canonical_name__ = "CodeHistoriesDict"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    code_versions: Dict[str, CodeHistoryView] = {}

    def _repr_html_(self):
        return f"""
            {self.code_versions._repr_html_()}
            """

    def add_func(self, versions: CodeHistoryView) -> Any:
        self.code_versions[versions.service_func_name] = versions

    def __getitem__(self, name: str) -> Any:
        return self.code_versions[name]

    def __getattr__(self, name: str) -> Any:
        code_versions = object.__getattribute__(self, "code_versions")
        if name in code_versions.keys():
            return code_versions[name]
        return object.__getattribute__(self, name)


@serializable()
class UsersCodeHistoriesDict(SyftObject):
    # version
    __canonical_name__ = "UsersCodeHistoriesDict"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    node_uid: UID
    user_dict: Dict[str, List[str]] = {}

    __repr_attrs__ = ["available_keys"]

    @property
    def available_keys(self):
        return json.dumps(self.user_dict, sort_keys=True, indent=4)

    def __getitem__(self, key: int):
        api = APIRegistry.api_for(self.node_uid, self.syft_client_verify_key)
        return api.services.code_history.get_history_for_user(key)

    def _repr_html_(self):
        rows = []
        for user, funcs in self.user_dict.items():
            rows += [{"user": user, "UserCodes": funcs}]
        return create_table_template(rows, "UserCodeHistory", table_icon=None)



# future
from __future__ import annotations

# stdlib
from typing import Optional

# third party
from typing_extensions import Self

# relative
from ..abstract_node import AbstractNode
from ..abstract_node import NodeSideType
from ..abstract_node import NodeType
from ..node.credentials import SyftSigningKey
from ..serde.serializable import serializable
from ..service.queue.base_queue import QueueConfig
from ..store.blob_storage import BlobStorageConfig
from ..store.document_store import StoreConfig
from ..types.syft_migration import migrate
from ..types.syft_object import SYFT_OBJECT_VERSION_1
from ..types.syft_object import SYFT_OBJECT_VERSION_2
from ..types.syft_object import SyftObject
from ..types.transforms import drop
from ..types.transforms import make_set_default
from ..types.uid import UID


@serializable()
class WorkerSettingsV1(SyftObject):
    __canonical_name__ = "WorkerSettings"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    name: str
    node_type: NodeType
    node_side_type: NodeSideType
    signing_key: SyftSigningKey
    document_store_config: StoreConfig
    action_store_config: StoreConfig
    blob_store_config: Optional[BlobStorageConfig]


@serializable()
class WorkerSettings(SyftObject):
    __canonical_name__ = "WorkerSettings"
    __version__ = SYFT_OBJECT_VERSION_2

    id: UID
    name: str
    node_type: NodeType
    node_side_type: NodeSideType
    signing_key: SyftSigningKey
    document_store_config: StoreConfig
    action_store_config: StoreConfig
    blob_store_config: Optional[BlobStorageConfig]
    queue_config: Optional[QueueConfig]

    @staticmethod
    def from_node(node: AbstractNode) -> Self:
        return WorkerSettings(
            id=node.id,
            name=node.name,
            node_type=node.node_type,
            signing_key=node.signing_key,
            document_store_config=node.document_store_config,
            action_store_config=node.action_store_config,
            node_side_type=node.node_side_type.value,
            blob_store_config=node.blob_store_config,
            queue_config=node.queue_config,
        )


# queue_config


@migrate(WorkerSettings, WorkerSettingsV1)
def downgrade_workersettings_v2_to_v1():
    return [
        drop(["queue_config"]),
    ]


@migrate(WorkerSettingsV1, WorkerSettings)
def upgrade_workersettings_v1_to_v2():
    # relative
    from ..service.queue.zmq_queue import ZMQQueueConfig

    return [make_set_default("queue_config", ZMQQueueConfig())]



# stdlib
from typing import Dict
import venv

# relative
from ..types.syft_object import SYFT_OBJECT_VERSION_1
from ..types.syft_object import SyftObject


class Env(SyftObject):
    __canonical_name__ = "Env"
    __version__ = SYFT_OBJECT_VERSION_1
    packages_dict: Dict[str, str]

    @property
    def packages(self) -> list[tuple[str, str]]:
        return list(self.packages_dict.items())

    def create_local_env(self) -> None:
        venv.EnvBuilder()



# Reference: https://github.com/pydantic/pydantic/issues/1223#issuecomment-998160737


# stdlib
import inspect
import threading
from typing import Any
from typing import Dict
from typing import Generator
from typing import Tuple
from typing import Type

# third party
from pydantic.fields import UndefinedType
from pydantic.main import BaseModel
from pydantic.main import ModelField
from pydantic.main import ModelMetaclass

# relative
from ..serde.recursive_primitives import recursive_serde_register_type
from ..serde.serializable import serializable

TupleGenerator = Generator[Tuple[str, Any], None, None]


@serializable()
class Empty:
    pass


class PartialModelMetaclass(ModelMetaclass):
    def __new__(
        meta: Type["PartialModelMetaclass"], *args: Any, **kwargs: Any
    ) -> "PartialModelMetaclass":
        cls = super().__new__(meta, *args, *kwargs)
        cls_init = cls.__init__
        # Because the class will be modified temporarily, need to lock __init__
        init_lock = threading.Lock()
        # To preserve identical hashes of temporary nested partial models,
        # only one instance of each temporary partial class can exist

        def __init__(self: BaseModel, *args: Any, **kwargs: Any) -> None:
            with init_lock:
                fields = self.__class__.__fields__
                fields_map: Dict[ModelField, Tuple[Any, bool]] = {}

                def optionalize(
                    fields: Dict[str, ModelField], *, restore: bool = False
                ) -> None:
                    for _, field in fields.items():
                        if not restore:
                            if isinstance(field.required, UndefinedType):
                                raise Exception(f"{field.name} is a required field.")
                            fields_map[field] = (field.type_, field.required)
                            # If field has None allowed as a value
                            # then it becomes a required field.
                            if field.allow_none and field.name in kwargs:
                                field.required = True
                            else:
                                field.required = False
                            if inspect.isclass(field.type_) and issubclass(
                                field.type_, BaseModel
                            ):
                                field.populate_validators()
                                if field.sub_fields is not None:
                                    for sub_field in field.sub_fields:
                                        sub_field.type_ = field.type_
                                        sub_field.populate_validators()
                                optionalize(field.type_.__fields__)
                        else:
                            # No need to recursively de-optionalize once original types
                            # are restored
                            field.type_, field.required = fields_map[field]
                            if field.sub_fields is not None:
                                for sub_field in field.sub_fields:
                                    sub_field.type_ = field.type_

                # Make fields and fields of nested model types optional
                optionalize(fields)

                # Transform kwargs that are PartialModels to their dict() forms. This
                # will exclude `None` (see below) from the dictionary used to construct
                # the temporarily-partial model field, avoiding ValidationErrors of
                # type type_error.none.not_allowed.
                for kwarg, value in kwargs.items():
                    if value.__class__.__class__ is PartialModelMetaclass:
                        kwargs[kwarg] = value.dict()
                    elif isinstance(value, (tuple, list)):
                        kwargs[kwarg] = value.__class__(
                            (
                                v.dict()
                                if v.__class__.__class__ is PartialModelMetaclass
                                else v
                            )
                            for v in value
                        )

                # Validation is performed in __init__, for which all fields are now optional
                cls_init(self, *args, **kwargs)
                # Restore requiredness
                optionalize(fields, restore=True)

        cls.__init__ = __init__

        def iter_exclude_empty(self) -> TupleGenerator:
            for key, value in self.__dict__.items():
                if value is not Empty:
                    yield key, value

        cls.__iter__ = iter_exclude_empty

        return cls


recursive_serde_register_type(PartialModelMetaclass)



# stdlib
from typing import Optional
from typing import Union

# third party
from typing_extensions import Self

# relative
from ...custom_worker.utils import ImageUtils
from ...serde.serializable import serializable
from ...types.base import SyftBaseModel
from .image_registry import SyftImageRegistry


@serializable()
class SyftWorkerImageIdentifier(SyftBaseModel):
    """
    Class to identify syft worker images.
    If a user provides an image's identifier with
    "docker.io/openmined/test-nginx:0.7.8", the convention we use for
    image name, tag and repo for now is
        tag = 0.7.8
        repo = openmined/test-nginx
        repo_with_tag = openmined/test-nginx:0.7.8
        full_name = docker.io/openmined/test-nginx
        full_name_with_tag = docker.io/openmined/test-nginx:0.7.8

    References:
        https://docs.docker.com/engine/reference/commandline/tag/#tag-an-image-referenced-by-name-and-tag
    """

    registry: Optional[Union[SyftImageRegistry, str]]
    repo: str
    tag: str

    __repr_attrs__ = ["registry", "repo", "tag"]

    @classmethod
    def with_registry(cls, tag: str, registry: SyftImageRegistry) -> Self:
        """Build a SyftWorkerImageTag from Docker tag & a previously created SyftImageRegistry object."""
        registry_str, repo, tag = ImageUtils.parse_tag(tag)

        # if we parsed a registry string, make sure it matches the registry object
        if registry_str and registry_str != registry.url:
            raise ValueError(f"Registry URL mismatch: {registry_str} != {registry.url}")

        return cls(repo=repo, tag=tag, registry=registry)

    @classmethod
    def from_str(cls, tag: str) -> Self:
        """Build a SyftWorkerImageTag from a pure-string standard Docker tag."""
        registry, repo, tag = ImageUtils.parse_tag(tag)
        return cls(repo=repo, registry=registry, tag=tag)

    @property
    def repo_with_tag(self) -> Optional[str]:
        if self.repo or self.tag:
            return f"{self.repo}:{self.tag}"
        return None

    @property
    def full_name_with_tag(self) -> str:
        if self.registry is None:
            return f"docker.io/{self.repo}:{self.tag}"
        elif isinstance(self.registry, str):
            return f"{self.registry}/{self.repo}:{self.tag}"
        else:
            return f"{self.registry.url}/{self.repo}:{self.tag}"

    @property
    def registry_host(self) -> str:
        if self.registry is None:
            return ""
        elif isinstance(self.registry, str):
            return self.registry
        else:
            return self.registry.url

    def __hash__(self) -> int:
        return hash(self.repo + self.tag + str(hash(self.registry)))

    def __str__(self) -> str:
        return self.full_name_with_tag

    def __repr__(self) -> str:
        return f"SyftWorkerImageIdentifier(repo={self.repo}, tag={self.tag}, registry={self.registry})"



# stdlib
from typing import Any
from typing import Callable
from typing import Dict
from typing import List

# relative
from ...serde.serializable import serializable
from ...store.document_store import SYFT_OBJECT_VERSION_1
from ...store.document_store import SyftObject
from ...types.datetime import DateTime
from ...types.syft_migration import migrate
from ...types.syft_object import SYFT_OBJECT_VERSION_2
from ...types.transforms import drop
from ...types.transforms import make_set_default


@serializable()
class DockerWorkerV1(SyftObject):
    # version
    __canonical_name__ = "ContainerImage"
    __version__ = SYFT_OBJECT_VERSION_1

    __attr_searchable__ = ["container_id"]
    __attr_unique__ = ["container_id"]
    __repr_attrs__ = ["container_id", "created_at"]

    container_id: str
    created_at: DateTime = DateTime.now()


@serializable()
class DockerWorker(SyftObject):
    # version
    __canonical_name__ = "ContainerImage"
    __version__ = SYFT_OBJECT_VERSION_2

    __attr_searchable__ = ["container_id", "container_name"]
    __attr_unique__ = ["container_id"]
    __repr_attrs__ = ["container_id", "created_at"]

    container_name: str
    container_id: str
    created_at: DateTime = DateTime.now()

    def _coll_repr_(self) -> Dict[str, Any]:
        return {
            "container_name": self.container_name,
            "container_id": self.container_id,
            "created_at": self.created_at,
        }


@migrate(DockerWorker, DockerWorkerV1)
def downgrade_job_v2_to_v1() -> List[Callable]:
    return [drop(["container_name"])]


@migrate(DockerWorkerV1, DockerWorker)
def upgrade_job_v2_to_v3() -> List[Callable]:
    return [make_set_default("job_consumer_id", None)]



# stdlib
import contextlib
import os
from pathlib import Path
import socket
import socketserver
import sys
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
import docker
from docker.models.containers import Container
from kr8s.objects import Pod

# relative
from ...abstract_node import AbstractNode
from ...custom_worker.builder import CustomWorkerBuilder
from ...custom_worker.builder_types import ImageBuildResult
from ...custom_worker.builder_types import ImagePushResult
from ...custom_worker.config import DockerWorkerConfig
from ...custom_worker.config import PrebuiltWorkerConfig
from ...custom_worker.k8s import KubeUtils
from ...custom_worker.k8s import PodStatus
from ...custom_worker.runner_k8s import KubernetesRunner
from ...node.credentials import SyftVerifyKey
from ...types.uid import UID
from ...util.util import get_queue_address
from ..response import SyftError
from .image_identifier import SyftWorkerImageIdentifier
from .worker_image import SyftWorkerImage
from .worker_image_stash import SyftWorkerImageStash
from .worker_pool import ContainerSpawnStatus
from .worker_pool import SyftWorker
from .worker_pool import WorkerHealth
from .worker_pool import WorkerOrchestrationType
from .worker_pool import WorkerStatus

DEFAULT_WORKER_IMAGE_TAG = "openmined/default-worker-image-cpu:0.0.1"
DEFAULT_WORKER_POOL_NAME = "default-pool"
K8S_NODE_CREDS_NAME = "node-creds"


def backend_container_name() -> str:
    hostname = socket.gethostname()
    service_name = os.getenv("SERVICE", "backend")
    return f"{hostname}-{service_name}-1"


def get_container(
    docker_client: docker.DockerClient, container_name: str
) -> Optional[Container]:
    try:
        existing_container = docker_client.containers.get(container_name)
    except docker.errors.NotFound:
        existing_container = None

    return existing_container


def extract_config_from_backend(
    worker_name: str, docker_client: docker.DockerClient
) -> Dict[str, Any]:
    # Existing main backend container
    backend_container = get_container(
        docker_client, container_name=backend_container_name()
    )

    # Config with defaults
    extracted_config: Dict[str, Any] = {
        "volume_binds": {},
        "network_mode": None,
        "environment": {},
    }

    if backend_container is None:
        return extracted_config

    # Inspect the existing container
    details = backend_container.attrs

    host_config = details["HostConfig"]
    environment = details["Config"]["Env"]

    # Extract Volume Binds
    for vol in host_config["Binds"]:
        parts = vol.split(":")
        key = parts[0]
        bind = parts[1]
        mode = parts[2]
        if "/storage" in bind:
            # we need this because otherwise we are using the same node private key
            # which will make account creation fail
            worker_postfix = worker_name.split("-", 1)[1]
            key = f"{key}-{worker_postfix}"
        extracted_config["volume_binds"][key] = {"bind": bind, "mode": mode}

    # Extract Environment Variables
    extracted_config["environment"] = dict([e.split("=", 1) for e in environment])
    extracted_config["network_mode"] = f"container:{backend_container.id}"

    return extracted_config


def get_free_tcp_port() -> int:
    with socketserver.TCPServer(("localhost", 0), None) as s:
        free_port = s.server_address[1]
        return free_port


def run_container_using_docker(
    worker_image: SyftWorkerImage,
    docker_client: docker.DockerClient,
    worker_name: str,
    worker_count: int,
    pool_name: str,
    queue_port: int,
    debug: bool = False,
    username: Optional[str] = None,
    password: Optional[str] = None,
    registry_url: Optional[str] = None,
) -> ContainerSpawnStatus:
    if not worker_image.is_built:
        raise ValueError("Image must be built before running it.")

    # Get hostname
    hostname = socket.gethostname()

    # Create a random UID for Worker
    syft_worker_uid = UID()

    # container name
    container_name = f"{hostname}-{worker_name}"

    # start container
    container = None
    error_message = None
    worker = None
    try:
        # login to the registry through the client
        # so that the subsequent pull/run commands work
        if registry_url and username and password:
            docker_client.login(
                username=username,
                password=password,
                registry=registry_url,
            )

        # If container with given name already exists then stop it
        # and recreate it.
        existing_container = get_container(
            docker_client=docker_client,
            container_name=container_name,
        )
        if existing_container:
            existing_container.remove(force=True)

        # Extract Config from backend container
        backend_host_config = extract_config_from_backend(
            worker_name=worker_name, docker_client=docker_client
        )

        # Create List of Envs to pass
        environment = backend_host_config["environment"]
        environment["CREATE_PRODUCER"] = "false"
        environment["N_CONSUMERS"] = 1
        environment["PORT"] = str(get_free_tcp_port())
        environment["HTTP_PORT"] = str(88 + worker_count)
        environment["HTTPS_PORT"] = str(446 + worker_count)
        environment["CONSUMER_SERVICE_NAME"] = pool_name
        environment["SYFT_WORKER_UID"] = syft_worker_uid
        environment["DEV_MODE"] = debug
        environment["QUEUE_PORT"] = queue_port
        environment["CONTAINER_HOST"] = "docker"

        if worker_image.image_identifier is None:
            raise ValueError(f"Image {worker_image} does not have an identifier")

        container = docker_client.containers.run(
            image=worker_image.image_identifier.full_name_with_tag,
            name=f"{hostname}-{worker_name}",
            detach=True,
            auto_remove=True,
            network_mode=backend_host_config["network_mode"],
            environment=environment,
            volumes=backend_host_config["volume_binds"],
            tty=True,
            stdin_open=True,
            labels={"orgs.openmined.syft": "this is a syft worker container"},
        )

        status = (
            WorkerStatus.STOPPED
            if container.status == "exited"
            else WorkerStatus.PENDING
        )

        healthcheck: WorkerHealth = _get_healthcheck_based_on_status(status)

        worker = SyftWorker(
            id=syft_worker_uid,
            name=worker_name,
            container_id=container.id,
            status=status,
            healthcheck=healthcheck,
            image=worker_image,
            worker_pool_name=pool_name,
        )
    except Exception as e:
        error_message = f"Failed to run command in container. {worker_name} {worker_image}. {e}. {sys.stderr}"
        if container:
            worker = SyftWorker(
                name=worker_name,
                container_id=container.id,
                status=WorkerStatus.STOPPED,
                healthcheck=WorkerHealth.UNHEALTHY,
                image=worker_image,
                worker_pool_name=pool_name,
            )
            container.stop()

    return ContainerSpawnStatus(
        worker_name=worker_name, worker=worker, error=error_message
    )


def run_workers_in_threads(
    node: AbstractNode,
    pool_name: str,
    number: int,
    start_idx: int = 0,
) -> List[ContainerSpawnStatus]:
    results = []

    for worker_count in range(start_idx + 1, number + 1):
        error = None
        worker_name = f"{pool_name}-{worker_count}"
        worker = SyftWorker(
            name=worker_name,
            status=WorkerStatus.RUNNING,
            worker_pool_name=pool_name,
            healthcheck=WorkerHealth.HEALTHY,
        )
        try:
            port = node.queue_config.client_config.queue_port
            address = get_queue_address(port)
            node.add_consumer_for_service(
                service_name=pool_name,
                syft_worker_id=worker.id,
                address=address,
            )
        except Exception as e:
            print(
                "Failed to start consumer for"
                f"pool={pool_name} worker={worker_name}. Error: {e}"
            )
            worker.status = WorkerStatus.STOPPED
            error = str(e)

        container_status = ContainerSpawnStatus(
            worker_name=worker_name,
            worker=worker,
            error=error,
        )

        results.append(container_status)

    return results


def prepare_kubernetes_pool_env(
    runner: KubernetesRunner, env_vars: dict
) -> Tuple[List, Dict]:
    # get current backend pod name
    backend_pod_name = os.getenv("K8S_POD_NAME")
    if not backend_pod_name:
        raise ValueError("Pod name not provided in environment variable")

    # get current backend's credentials path
    creds_path: Optional[Union[str, Path]] = os.getenv("CREDENTIALS_PATH")
    if not creds_path:
        raise ValueError("Credentials path not provided")

    creds_path = Path(creds_path)
    if creds_path is not None and not creds_path.exists():
        raise ValueError("Credentials file does not exist")

    # create a secret for the node credentials owned by the backend, not the pool.
    node_secret = KubeUtils.create_secret(
        secret_name=K8S_NODE_CREDS_NAME,
        type="Opaque",
        component=backend_pod_name,
        data={creds_path.name: creds_path.read_text()},
        encoded=False,
    )

    # clone and patch backend environment variables
    backend_env = runner.get_pod_env_vars(backend_pod_name) or []
    env_vars_: List = KubeUtils.patch_env_vars(backend_env, env_vars)
    mount_secrets = {
        node_secret.metadata.name: {
            "mountPath": str(creds_path),
            "subPath": creds_path.name,
        },
    }

    return env_vars_, mount_secrets


def create_kubernetes_pool(
    runner: KubernetesRunner,
    tag: str,
    pool_name: str,
    replicas: int,
    queue_port: int,
    debug: bool,
    reg_username: Optional[str] = None,
    reg_password: Optional[str] = None,
    reg_url: Optional[str] = None,
    **kwargs: Any,
) -> Union[List[Pod], SyftError]:
    pool = None
    error = False

    try:
        print(
            "Creating new pool "
            f"name={pool_name} "
            f"tag={tag} "
            f"replicas={replicas}"
        )

        env_vars, mount_secrets = prepare_kubernetes_pool_env(
            runner,
            {
                "SYFT_WORKER": "True",
                "DEV_MODE": f"{debug}",
                "QUEUE_PORT": f"{queue_port}",
                "CONSUMER_SERVICE_NAME": pool_name,
                "N_CONSUMERS": "1",
                "CREATE_PRODUCER": "False",
                "INMEMORY_WORKERS": "False",
            },
        )

        # run the pool with args + secret
        pool = runner.create_pool(
            pool_name=pool_name,
            tag=tag,
            replicas=replicas,
            env_vars=env_vars,
            mount_secrets=mount_secrets,
            reg_username=reg_username,
            reg_password=reg_password,
            reg_url=reg_url,
        )
    except Exception as e:
        error = True
        return SyftError(message=f"Failed to start workers {e}")
    finally:
        if error and pool:
            pool.delete()

    return runner.get_pool_pods(pool_name=pool_name)


def scale_kubernetes_pool(
    runner: KubernetesRunner,
    pool_name: str,
    replicas: int,
) -> Union[List[Pod], SyftError]:
    pool = runner.get_pool(pool_name)
    if not pool:
        return SyftError(message=f"Pool does not exist. name={pool_name}")

    try:
        print(f"Scaling pool name={pool_name} to replicas={replicas}")
        runner.scale_pool(pool_name=pool_name, replicas=replicas)
    except Exception as e:
        return SyftError(message=f"Failed to scale workers {e}")

    return runner.get_pool_pods(pool_name=pool_name)


def run_workers_in_kubernetes(
    worker_image: SyftWorkerImage,
    worker_count: int,
    pool_name: str,
    queue_port: int,
    start_idx: int = 0,
    debug: bool = False,
    reg_username: Optional[str] = None,
    reg_password: Optional[str] = None,
    reg_url: Optional[str] = None,
    **kwargs: Any,
) -> Union[List[ContainerSpawnStatus], SyftError]:
    spawn_status = []
    runner = KubernetesRunner()

    if not runner.exists(pool_name=pool_name):
        if worker_image.image_identifier is not None:
            pool_pods = create_kubernetes_pool(
                runner=runner,
                tag=worker_image.image_identifier.full_name_with_tag,
                pool_name=pool_name,
                replicas=worker_count,
                queue_port=queue_port,
                debug=debug,
                reg_username=reg_username,
                reg_password=reg_password,
                reg_url=reg_url,
            )
        else:
            return SyftError(
                message=f"image with uid {worker_image.id} does not have an image identifier"
            )
    else:
        pool_pods = scale_kubernetes_pool(runner, pool_name, worker_count)

        if isinstance(pool_pods, list) and len(pool_pods) > 0:
            # slice only those pods that we're interested in
            pool_pods = pool_pods[start_idx:]

    if isinstance(pool_pods, SyftError):
        return pool_pods

    # create worker object
    for pod in pool_pods:
        status = runner.get_pod_status(pod)
        status, healthcheck, error = map_pod_to_worker_status(status)

        # this worker id will be the same as the one in the worker
        syft_worker_uid = UID.with_seed(pod.metadata.name)

        worker = SyftWorker(
            id=syft_worker_uid,
            name=pod.metadata.name,
            container_id=None,
            status=status,
            healthcheck=healthcheck,
            image=worker_image,
            worker_pool_name=pool_name,
        )

        spawn_status.append(
            ContainerSpawnStatus(
                worker_name=pod.metadata.name,
                worker=worker,
                error=error,
            )
        )

    return spawn_status


def map_pod_to_worker_status(
    status: PodStatus,
) -> Tuple[WorkerStatus, WorkerHealth, Optional[str]]:
    worker_status = None
    worker_healthcheck = None
    worker_error = None

    # check if pod is ready through pod.status.condition.Ready & pod.status.condition.ContainersReady
    pod_ready = status.condition.ready and status.condition.containers_ready

    if not pod_ready:
        # extract error if not ready
        worker_error = f"{status.container.reason}: {status.container.message}"

    # map readiness to status - it's either running or pending.
    # closely relates to pod.status.phase, but avoiding as it is not as detailed as pod.status.conditions
    worker_status = WorkerStatus.RUNNING if pod_ready else WorkerStatus.PENDING

    # TODO: update these values based on actual runtime probes instead of kube pod statuses
    # if there are any errors, then healthcheck is unhealthy
    worker_healthcheck = (
        WorkerHealth.UNHEALTHY if worker_error else WorkerHealth.HEALTHY
    )

    return worker_status, worker_healthcheck, worker_error


def run_containers(
    pool_name: str,
    worker_image: SyftWorkerImage,
    number: int,
    orchestration: WorkerOrchestrationType,
    queue_port: int,
    dev_mode: bool = False,
    start_idx: int = 0,
    reg_username: Optional[str] = None,
    reg_password: Optional[str] = None,
    reg_url: Optional[str] = None,
) -> Union[List[ContainerSpawnStatus], SyftError]:
    results = []

    if not worker_image.is_built:
        return SyftError(message="Image must be built before running it.")

    print(f"Starting workers with start_idx={start_idx} count={number}")

    if orchestration == WorkerOrchestrationType.DOCKER:
        with contextlib.closing(docker.from_env()) as client:
            for worker_count in range(start_idx + 1, number + 1):
                worker_name = f"{pool_name}-{worker_count}"
                spawn_result = run_container_using_docker(
                    docker_client=client,
                    worker_name=worker_name,
                    worker_count=worker_count,
                    worker_image=worker_image,
                    pool_name=pool_name,
                    queue_port=queue_port,
                    debug=dev_mode,
                    username=reg_username,
                    password=reg_password,
                    registry_url=reg_url,
                )
                results.append(spawn_result)
    elif orchestration == WorkerOrchestrationType.KUBERNETES:
        return run_workers_in_kubernetes(
            worker_image=worker_image,
            worker_count=number,
            pool_name=pool_name,
            queue_port=queue_port,
            debug=dev_mode,
            start_idx=start_idx,
            reg_username=reg_username,
            reg_password=reg_password,
            reg_url=reg_url,
        )

    return results


def create_default_image(
    credentials: SyftVerifyKey,
    image_stash: SyftWorkerImageStash,
    tag: str,
    in_kubernetes: bool = False,
) -> Union[SyftError, SyftWorkerImage]:
    # TODO: Hardcode worker dockerfile since not able to COPY
    # worker_cpu.dockerfile to backend in backend.dockerfile.

    # default_cpu_dockerfile = get_syft_cpu_dockerfile()
    # DockerWorkerConfig.from_path(default_cpu_dockerfile)

    if not in_kubernetes:
        default_cpu_dockerfile = f"""ARG SYFT_VERSION_TAG='{tag}' \n"""
        default_cpu_dockerfile += """FROM openmined/grid-backend:${SYFT_VERSION_TAG}
        ARG PYTHON_VERSION="3.11"
        ARG SYSTEM_PACKAGES=""
        ARG PIP_PACKAGES="pip --dry-run"
        ARG CUSTOM_CMD='echo "No custom commands passed"'

        # Worker specific environment variables go here
        ENV SYFT_WORKER="true"
        ENV DOCKER_TAG=${SYFT_VERSION_TAG}

        RUN apk update && \
            apk add ${SYSTEM_PACKAGES} && \
            pip install --user ${PIP_PACKAGES} && \
            bash -c "$CUSTOM_CMD"
        """
        worker_config = DockerWorkerConfig(dockerfile=default_cpu_dockerfile)
        _new_image = SyftWorkerImage(
            config=worker_config,
            created_by=credentials,
        )
    else:
        # in k8s we don't need to build the image, just the tag of backend is enough
        worker_config = PrebuiltWorkerConfig(
            tag=tag,
            description="Prebuilt default worker image",
        )

        # create SyftWorkerImage from a pre-built image
        _new_image = SyftWorkerImage(
            config=worker_config,
            created_by=credentials,
            image_identifier=SyftWorkerImageIdentifier.from_str(tag),
        )

    result = image_stash.get_by_docker_config(
        credentials=credentials,
        config=worker_config,
    )

    if result.ok() is None:
        result = image_stash.set(credentials, _new_image)
        if result.is_err():
            return SyftError(message=f"Failed to save image stash: {result.err()}")

    default_syft_image = result.ok()

    return default_syft_image


def _get_healthcheck_based_on_status(status: WorkerStatus) -> WorkerHealth:
    if status in [WorkerStatus.PENDING, WorkerStatus.RUNNING]:
        return WorkerHealth.HEALTHY
    else:
        return WorkerHealth.UNHEALTHY


def image_build(
    image: SyftWorkerImage, **kwargs: Dict[str, Any]
) -> Union[ImageBuildResult, SyftError]:
    if image.image_identifier is not None:
        full_tag = image.image_identifier.full_name_with_tag
        try:
            builder = CustomWorkerBuilder()
            return builder.build_image(
                config=image.config,
                tag=full_tag,
                rm=True,
                forcerm=True,
                **kwargs,
            )
        except docker.errors.APIError as e:
            return SyftError(
                message=f"Docker API error when building '{full_tag}'. Reason - {e}"
            )
        except docker.errors.DockerException as e:
            return SyftError(
                message=f"Docker exception when building '{full_tag}'. Reason - {e}"
            )
        except Exception as e:
            return SyftError(
                message=f"Unknown exception when building '{full_tag}'. Reason - {e}"
            )
    else:
        return SyftError(
            message=f"image with uid {image.id} does not have an image identifier"
        )


def image_push(
    image: SyftWorkerImage,
    username: Optional[str] = None,
    password: Optional[str] = None,
) -> Union[ImagePushResult, SyftError]:
    if image.image_identifier is not None:
        full_tag = image.image_identifier.full_name_with_tag
        try:
            builder = CustomWorkerBuilder()
            result = builder.push_image(
                # this should be consistent with docker build command
                tag=image.image_identifier.full_name_with_tag,
                registry_url=image.image_identifier.registry_host,
                username=username,
                password=password,
            )

            if "error" in result.logs.lower() or result.exit_code:
                return SyftError(
                    message=f"Failed to push {full_tag}. "
                    f"Exit code: {result.exit_code}. "
                    f"Logs:\n{result.logs}"
                )

            return result
        except docker.errors.APIError as e:
            return SyftError(message=f"Docker API error when pushing {full_tag}. {e}")
        except docker.errors.DockerException as e:
            return SyftError(
                message=f"Docker exception when pushing {full_tag}. Reason - {e}"
            )
        except Exception as e:
            return SyftError(
                message=f"Unknown exception when pushing {image.image_identifier}. Reason - {e}"
            )
    else:
        return SyftError(
            message=f"image with uid {image.id} does not have an "
            "image identifier and tag, hence we can't push it."
        )


def get_orchestration_type() -> WorkerOrchestrationType:
    """Returns orchestration type from env. Defaults to Python."""

    orchstration_type_ = os.getenv("CONTAINER_HOST")
    return (
        WorkerOrchestrationType(orchstration_type_.lower())
        if orchstration_type_
        else WorkerOrchestrationType.PYTHON
    )



# stdlib
from binascii import hexlify
from collections import defaultdict
import itertools
import socketserver
import threading
import time
from time import sleep
from typing import DefaultDict
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

# third party
from loguru import logger
from pydantic import validator
from zmq import Frame
from zmq import LINGER
from zmq.error import ContextTerminated
import zmq.green as zmq

# relative
from ...serde.deserialize import _deserialize
from ...serde.serializable import serializable
from ...serde.serialize import _serialize as serialize
from ...service.action.action_object import ActionObject
from ...service.context import AuthedServiceContext
from ...types.base import SyftBaseModel
from ...types.syft_migration import migrate
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SYFT_OBJECT_VERSION_2
from ...types.syft_object import SYFT_OBJECT_VERSION_3
from ...types.syft_object import SyftObject
from ...types.transforms import drop
from ...types.transforms import make_set_default
from ...types.uid import UID
from ...util.util import get_queue_address
from ..response import SyftError
from ..response import SyftSuccess
from ..worker.worker_pool import ConsumerState
from ..worker.worker_stash import WorkerStash
from .base_queue import AbstractMessageHandler
from .base_queue import QueueClient
from .base_queue import QueueClientConfig
from .base_queue import QueueConfig
from .base_queue import QueueConsumer
from .base_queue import QueueProducer
from .queue_stash import ActionQueueItem
from .queue_stash import Status

# Producer/Consumer heartbeat interval (in seconds)
HEARTBEAT_INTERVAL_SEC = 2

# Thread join timeout (in seconds)
THREAD_TIMEOUT_SEC = 5

# Max duration (in ms) to wait for ZMQ poller to return
ZMQ_POLLER_TIMEOUT_MSEC = 1000

# Duration (in seconds) after which a worker without a heartbeat will be marked as expired
WORKER_TIMEOUT_SEC = 60

# Duration (in seconds) after which producer without a heartbeat will be marked as expired
PRODUCER_TIMEOUT_SEC = 60

# Lock for working on ZMQ socket
ZMQ_SOCKET_LOCK = threading.Lock()


class QueueMsgProtocol:
    W_WORKER = b"MDPW01"
    W_READY = b"0x01"
    W_REQUEST = b"0x02"
    W_REPLY = b"0x03"
    W_HEARTBEAT = b"0x04"
    W_DISCONNECT = b"0x05"


MAX_RECURSION_NESTED_ACTIONOBJECTS = 5


class Timeout:
    def __init__(self, offset_sec: float):
        self.__offset = float(offset_sec)
        self.__next_ts = 0

        self.reset()

    @property
    def next_ts(self):
        return self.__next_ts

    def reset(self):
        self.__next_ts = self.now() + self.__offset

    def has_expired(self):
        return self.now() >= self.__next_ts

    @staticmethod
    def now() -> float:
        return time.time()


class Worker(SyftBaseModel):
    address: bytes
    identity: bytes
    service: Optional[str] = None
    syft_worker_id: Optional[UID] = None
    expiry_t: Timeout = Timeout(WORKER_TIMEOUT_SEC)

    @validator("syft_worker_id", pre=True, always=True)
    def set_syft_worker_id(cls, v, values):
        if isinstance(v, str):
            return UID(v)
        return v

    def has_expired(self):
        return self.expiry_t.has_expired()

    def get_expiry(self) -> int:
        return self.expiry_t.next_ts

    def reset_expiry(self):
        self.expiry_t.reset()


class Service:
    def __init__(self, name: str) -> None:
        self.name = name
        self.requests = []
        self.waiting = []  # List of waiting workers


@serializable()
class ZMQProducer(QueueProducer):
    INTERNAL_SERVICE_PREFIX = b"mmi."

    def __init__(
        self,
        queue_name: str,
        queue_stash,
        worker_stash: WorkerStash,
        port: int,
        context: AuthedServiceContext,
    ) -> None:
        self.id = UID().short()
        self.port = port
        self.queue_stash = queue_stash
        self.worker_stash = worker_stash
        self.queue_name = queue_name
        self.auth_context = context
        self._stop = threading.Event()
        self.post_init()

    @property
    def address(self):
        return get_queue_address(self.port)

    def post_init(self):
        """Initialize producer state."""

        self.services = {}
        self.workers = {}
        self.waiting: List[Worker] = []
        self.heartbeat_t = Timeout(HEARTBEAT_INTERVAL_SEC)
        self.context = zmq.Context(1)
        self.socket = self.context.socket(zmq.ROUTER)
        self.socket.setsockopt(LINGER, 1)
        self.socket.setsockopt_string(zmq.IDENTITY, self.id)
        self.poll_workers = zmq.Poller()
        self.poll_workers.register(self.socket, zmq.POLLIN)
        self.bind(f"tcp://*:{self.port}")
        self.thread: threading.Thread = None
        self.producer_thread: threading.Thread = None

    def close(self):
        self._stop.set()

        try:
            self.poll_workers.unregister(self.socket)
        except Exception as e:
            logger.exception("Failed to unregister poller. {}", e)
        finally:
            if self.thread:
                self.thread.join(THREAD_TIMEOUT_SEC)
                self.thread = None

            if self.producer_thread:
                self.producer_thread.join(THREAD_TIMEOUT_SEC)
                self.producer_thread = None

            self.socket.close()
            self.context.destroy()

            self._stop.clear()

    @property
    def action_service(self):
        return self.auth_context.node.get_service("ActionService")

    def contains_unresolved_action_objects(self, arg, recursion=0):
        """recursively check collections for unresolved action objects"""
        if isinstance(arg, UID):
            arg = self.action_service.get(self.auth_context, arg).ok()
            return self.contains_unresolved_action_objects(arg, recursion=recursion + 1)
        if isinstance(arg, ActionObject):
            if not arg.syft_resolved:
                res = self.action_service.get(self.auth_context, arg)
                if res.is_err():
                    return True
                arg = res.ok()
                if not arg.syft_resolved:
                    return True
            arg = arg.syft_action_data

        try:
            value = False
            if isinstance(arg, List):
                for elem in arg:
                    value = self.contains_unresolved_action_objects(
                        elem, recursion=recursion + 1
                    )
                    if value:
                        return True
            if isinstance(arg, Dict):
                for elem in arg.values():
                    value = self.contains_unresolved_action_objects(
                        elem, recursion=recursion + 1
                    )
                    if value:
                        return True
            return value
        except Exception as e:
            logger.exception("Failed to resolve action objects. {}", e)
            return True

    def unwrap_nested_actionobjects(self, data):
        """recursively unwraps nested action objects"""

        if isinstance(data, List):
            return [self.unwrap_nested_actionobjects(obj) for obj in data]
        if isinstance(data, Dict):
            return {
                key: self.unwrap_nested_actionobjects(obj) for key, obj in data.items()
            }
        if isinstance(data, ActionObject):
            res = self.action_service.get(self.auth_context, data.id)
            res = res.ok() if res.is_ok() else res.err()
            if not isinstance(res, ActionObject):
                return SyftError(message=f"{res}")
            else:
                nested_res = res.syft_action_data
                if isinstance(nested_res, ActionObject):
                    nested_res.syft_node_location = res.syft_node_location
                    nested_res.syft_client_verify_key = res.syft_client_verify_key
                return nested_res
        return data

    def preprocess_action_arg(self, arg):
        res = self.action_service.get(context=self.auth_context, uid=arg)
        if res.is_err():
            return arg
        action_object = res.ok()
        data = action_object.syft_action_data
        new_data = self.unwrap_nested_actionobjects(data)
        new_action_object = ActionObject.from_obj(new_data, id=action_object.id)
        res = self.action_service.set(
            context=self.auth_context, action_object=new_action_object
        )

    def read_items(self):
        while True:
            if self._stop.is_set():
                break
            sleep(1)

            # Items to be queued
            items_to_queue = self.queue_stash.get_by_status(
                self.queue_stash.partition.root_verify_key,
                status=Status.CREATED,
            ).ok()

            items_to_queue = [] if items_to_queue is None else items_to_queue

            # Queue Items that are in the processing state
            items_processing = self.queue_stash.get_by_status(
                self.queue_stash.partition.root_verify_key,
                status=Status.PROCESSING,
            ).ok()

            items_processing = [] if items_processing is None else items_processing

            for item in itertools.chain(items_to_queue, items_processing):
                if item.status == Status.CREATED:
                    if isinstance(item, ActionQueueItem):
                        action = item.kwargs["action"]
                        if self.contains_unresolved_action_objects(
                            action.args
                        ) or self.contains_unresolved_action_objects(action.kwargs):
                            continue
                        for arg in action.args:
                            self.preprocess_action_arg(arg)
                        for _, arg in action.kwargs.items():
                            self.preprocess_action_arg(arg)

                    msg_bytes = serialize(item, to_bytes=True)
                    worker_pool = item.worker_pool.resolve_with_context(
                        self.auth_context
                    )
                    worker_pool = worker_pool.ok()
                    service_name = worker_pool.name
                    service: Service = self.services.get(service_name)

                    # Skip adding message if corresponding service/pool
                    # is not registered.
                    if service is None:
                        continue

                    # append request message to the corresponding service
                    # This list is processed in dispatch method.

                    # TODO: Logic to evaluate the CAN RUN Condition
                    service.requests.append(msg_bytes)
                    item.status = Status.PROCESSING
                    res = self.queue_stash.update(item.syft_client_verify_key, item)
                    if res.is_err():
                        logger.error(
                            "Failed to update queue item={} error={}",
                            item,
                            res.err(),
                        )
                elif item.status == Status.PROCESSING:
                    # Evaluate Retry condition here
                    # If job running and timeout or job status is KILL
                    # or heartbeat fails
                    # or container id doesn't exists, kill process or container
                    # else decrease retry count and mark status as CREATED.
                    pass

    def run(self):
        self.thread = threading.Thread(target=self._run)
        self.thread.start()

        self.producer_thread = threading.Thread(target=self.read_items)
        self.producer_thread.start()

    def send(self, worker: bytes, message: Union[bytes, List[bytes]]):
        worker_obj = self.require_worker(worker)
        self.send_to_worker(worker=worker_obj, msg=message)

    def bind(self, endpoint):
        """Bind producer to endpoint."""
        self.socket.bind(endpoint)
        logger.info("Producer endpoint: {}", endpoint)

    def send_heartbeats(self):
        """Send heartbeats to idle workers if it's time"""
        if self.heartbeat_t.has_expired():
            for worker in self.waiting:
                self.send_to_worker(worker, QueueMsgProtocol.W_HEARTBEAT, None, None)
            self.heartbeat_t.reset()

    def purge_workers(self):
        """Look for & kill expired workers.

        Workers are oldest to most recent, so we stop at the first alive worker.
        """
        # work on a copy of the iterator
        for worker in list(self.waiting):
            if worker.has_expired():
                logger.info(
                    "Deleting expired Worker id={} uid={} expiry={} now={}",
                    worker.identity,
                    worker.syft_worker_id,
                    worker.get_expiry(),
                    Timeout.now(),
                )
                self.delete_worker(worker, False)

    def update_consumer_state_for_worker(
        self, syft_worker_id: UID, consumer_state: ConsumerState
    ):
        if self.worker_stash is None:
            logger.error(
                f"Worker stash is not defined for ZMQProducer : {self.queue_name} - {self.id}"
            )
            return

        try:
            res = self.worker_stash.update_consumer_state(
                credentials=self.worker_stash.partition.root_verify_key,
                worker_uid=syft_worker_id,
                consumer_state=consumer_state,
            )
            if res.is_err():
                logger.error(
                    "Failed to update consumer state for worker id={} error={}",
                    syft_worker_id,
                    res.err(),
                )
        except Exception as e:
            logger.error(
                f"Failed to update consumer state for worker id: {syft_worker_id}. Error: {e}"
            )

    def worker_waiting(self, worker: Worker):
        """This worker is now waiting for work."""
        # Queue to broker and service waiting lists
        if worker not in self.waiting:
            self.waiting.append(worker)
        if worker not in worker.service.waiting:
            worker.service.waiting.append(worker)
        worker.reset_expiry()
        self.update_consumer_state_for_worker(worker.syft_worker_id, ConsumerState.IDLE)
        self.dispatch(worker.service, None)

    def dispatch(self, service: Service, msg: bytes):
        """Dispatch requests to waiting workers as possible"""
        if msg is not None:  # Queue message if any
            service.requests.append(msg)

        self.purge_workers()
        while service.waiting and service.requests:
            # One worker consuming only one message at a time.
            msg = service.requests.pop(0)
            worker = service.waiting.pop(0)
            self.waiting.remove(worker)
            self.send_to_worker(worker, QueueMsgProtocol.W_REQUEST, None, msg)

    def send_to_worker(
        self,
        worker: Worker,
        command: QueueMsgProtocol = QueueMsgProtocol.W_REQUEST,
        option: bytes = None,
        msg: Optional[Union[bytes, list]] = None,
    ):
        """Send message to worker.

        If message is provided, sends that message.
        """

        if msg is None:
            msg = []
        elif not isinstance(msg, list):
            msg = [msg]

        # Stack routing and protocol envelopes to start of message
        # and routing envelope
        if option is not None:
            msg = [option] + msg
        msg = [worker.address, b"", QueueMsgProtocol.W_WORKER, command] + msg

        logger.debug("Send: {}", msg)
        with ZMQ_SOCKET_LOCK:
            self.socket.send_multipart(msg)

    def _run(self):
        while True:
            if self._stop.is_set():
                return

            for _, service in self.services.items():
                self.dispatch(service, None)

            items = None

            try:
                items = self.poll_workers.poll(ZMQ_POLLER_TIMEOUT_MSEC)
            except Exception as e:
                logger.exception("Failed to poll items: {}", e)

            if items:
                msg = self.socket.recv_multipart()

                logger.debug("Recieve: {}", msg)

                address = msg.pop(0)
                empty = msg.pop(0)  # noqa: F841
                header = msg.pop(0)

                if header == QueueMsgProtocol.W_WORKER:
                    self.process_worker(address, msg)
                else:
                    logger.error("Invalid message header: {}", header)

            self.send_heartbeats()
            self.purge_workers()

    def require_worker(self, address):
        """Finds the worker (creates if necessary)."""
        identity = hexlify(address)
        worker = self.workers.get(identity)
        if worker is None:
            worker = Worker(identity=identity, address=address)
            self.workers[identity] = worker
        return worker

    def process_worker(self, address: bytes, msg: List[bytes]):
        command = msg.pop(0)

        worker_ready = hexlify(address) in self.workers

        worker = self.require_worker(address)

        if QueueMsgProtocol.W_READY == command:
            service_name = msg.pop(0).decode()
            syft_worker_id = msg.pop(0).decode()
            if worker_ready:
                # Not first command in session or Reserved service name
                # If worker was already present, then we disconnect it first
                # and wait for it to re-register itself to the producer. This ensures that
                # we always have a healthy worker in place that can talk to the producer.
                self.delete_worker(worker, True)
            else:
                # Attach worker to service and mark as idle
                if service_name not in self.services:
                    service = Service(service_name)
                    self.services[service_name] = service
                else:
                    service = self.services.get(service_name)
                worker.service = service
                worker.syft_worker_id = UID(syft_worker_id)
                logger.info(
                    "New Worker service={} id={} uid={}",
                    service.name,
                    worker.identity,
                    worker.syft_worker_id,
                )
                self.worker_waiting(worker)

        elif QueueMsgProtocol.W_HEARTBEAT == command:
            if worker_ready:
                # If worker is ready then reset expiry
                # and add it to worker waiting list
                # if not already present
                self.worker_waiting(worker)
            else:
                # extract the syft worker id and worker pool name from the message
                # Get the corresponding worker pool and worker
                # update the status to be unhealthy
                self.delete_worker(worker, True)
        elif QueueMsgProtocol.W_DISCONNECT == command:
            self.delete_worker(worker, False)
        else:
            logger.error("Invalid command: {}", command)

    def delete_worker(self, worker: Worker, disconnect: bool):
        """Deletes worker from all data structures, and deletes worker."""
        if disconnect:
            self.send_to_worker(worker, QueueMsgProtocol.W_DISCONNECT, None, None)

        if worker.service and worker in worker.service.waiting:
            worker.service.waiting.remove(worker)

        if worker in self.waiting:
            self.waiting.remove(worker)

        self.workers.pop(worker.identity, None)

        self.update_consumer_state_for_worker(
            worker.syft_worker_id, ConsumerState.DETACHED
        )

    @property
    def alive(self):
        return not self.socket.closed


@serializable(attrs=["_subscriber"])
class ZMQConsumer(QueueConsumer):
    def __init__(
        self,
        message_handler: AbstractMessageHandler,
        address: str,
        queue_name: str,
        service_name: str,
        syft_worker_id: Optional[UID] = None,
        worker_stash: Optional[WorkerStash] = None,
        verbose: bool = False,
    ) -> None:
        self.address = address
        self.message_handler = message_handler
        self.service_name = service_name
        self.queue_name = queue_name
        self.context = zmq.Context()
        self.poller = zmq.Poller()
        self.socket = None
        self.verbose = verbose
        self.id = UID().short()
        self._stop = threading.Event()
        self.syft_worker_id = syft_worker_id
        self.worker_stash = worker_stash
        self.post_init()

    def reconnect_to_producer(self):
        """Connect or reconnect to producer"""
        if self.socket:
            self.poller.unregister(self.socket)
            self.socket.close()
        self.socket = self.context.socket(zmq.DEALER)
        self.socket.linger = 0
        self.socket.setsockopt_string(zmq.IDENTITY, self.id)
        self.socket.connect(self.address)
        self.poller.register(self.socket, zmq.POLLIN)

        logger.info("Connecting Worker id={} to broker addr={}", self.id, self.address)

        # Register queue with the producer
        self.send_to_producer(
            QueueMsgProtocol.W_READY,
            self.service_name.encode(),
            [str(self.syft_worker_id).encode()],
        )

    def post_init(self):
        self.thread = None
        self.heartbeat_t = Timeout(HEARTBEAT_INTERVAL_SEC)
        self.producer_ping_t = Timeout(PRODUCER_TIMEOUT_SEC)
        self.reconnect_to_producer()

    def close(self):
        self._stop.set()
        try:
            self.poller.unregister(self.socket)
        except Exception as e:
            logger.exception("Failed to unregister worker. {}", e)
        finally:
            if self.thread is not None:
                self.thread.join(timeout=THREAD_TIMEOUT_SEC)
                self.thread = None
            self.socket.close()
            self.context.destroy()
            self._stop.clear()

    def send_to_producer(
        self,
        command: str,
        option: Optional[bytes] = None,
        msg: Optional[Union[bytes, list]] = None,
    ):
        """Send message to producer.

        If no msg is provided, creates one internally
        """
        if msg is None:
            msg = []
        elif not isinstance(msg, list):
            msg = [msg]

        if option:
            msg = [option] + msg

        msg = [b"", QueueMsgProtocol.W_WORKER, command] + msg
        logger.debug("Send: msg={}", msg)
        with ZMQ_SOCKET_LOCK:
            self.socket.send_multipart(msg)

    def _run(self):
        """Send reply, if any, to producer and wait for next request."""
        try:
            while True:
                if self._stop.is_set():
                    return

                try:
                    items = self.poller.poll(ZMQ_POLLER_TIMEOUT_MSEC)
                except ContextTerminated:
                    logger.info("Context terminated")
                    return
                except Exception as e:
                    logger.error("Poll error={}", e)
                    continue

                if items:
                    # Message format:
                    # [b"", "<header>", "<command>", "<queue_name>", "<actual_msg_bytes>"]
                    msg = self.socket.recv_multipart()

                    logger.debug("Recieve: {}", msg)

                    # mark as alive
                    self.set_producer_alive()

                    if len(msg) < 3:
                        logger.error("Invalid message: {}", msg)
                        continue

                    empty = msg.pop(0)  # noqa: F841
                    header = msg.pop(0)  # noqa: F841

                    command = msg.pop(0)

                    if command == QueueMsgProtocol.W_REQUEST:
                        # Call Message Handler
                        try:
                            message = msg.pop()
                            self.associate_job(message)
                            self.message_handler.handle_message(
                                message=message,
                                syft_worker_id=self.syft_worker_id,
                            )
                        except Exception as e:
                            logger.exception("Error while handling message. {}", e)
                        finally:
                            self.clear_job()
                    elif command == QueueMsgProtocol.W_HEARTBEAT:
                        self.set_producer_alive()
                    elif command == QueueMsgProtocol.W_DISCONNECT:
                        self.reconnect_to_producer()
                    else:
                        logger.error("Invalid command: {}", command)
                else:
                    if not self.is_producer_alive():
                        logger.info("Producer check-alive timed out. Reconnecting.")
                        self.reconnect_to_producer()
                        self.set_producer_alive()

                self.send_heartbeat()

        except zmq.ZMQError as e:
            if e.errno == zmq.ETERM:
                logger.info("Consumer connection terminated")
            else:
                logger.exception("Consumer error. {}", e)
                raise e

        logger.info("Worker finished")

    def set_producer_alive(self):
        self.producer_ping_t.reset()

    def is_producer_alive(self) -> bool:
        # producer timer is within timeout
        return not self.producer_ping_t.has_expired()

    def send_heartbeat(self):
        if self.heartbeat_t.has_expired() and self.is_producer_alive():
            self.send_to_producer(QueueMsgProtocol.W_HEARTBEAT)
            self.heartbeat_t.reset()

    def run(self):
        self.thread = threading.Thread(target=self._run)
        self.thread.start()

    def associate_job(self, message: Frame):
        try:
            queue_item = _deserialize(message, from_bytes=True)
            self._set_worker_job(queue_item.job_id)
        except Exception as e:
            logger.exception("Could not associate job. {}", e)

    def clear_job(self):
        self._set_worker_job(None)

    def _set_worker_job(self, job_id: Optional[UID]):
        if self.worker_stash is not None:
            consumer_state = (
                ConsumerState.IDLE if job_id is None else ConsumerState.CONSUMING
            )
            res = self.worker_stash.update_consumer_state(
                credentials=self.worker_stash.partition.root_verify_key,
                worker_uid=self.syft_worker_id,
                consumer_state=consumer_state,
            )
            if res.is_err():
                logger.error(
                    f"Failed to update consumer state for {self.service_name}-{self.id}, error={res.err()}"
                )

    @property
    def alive(self):
        return not self.socket.closed and self.is_producer_alive()


@serializable()
class ZMQClientConfigV1(SyftObject, QueueClientConfig):
    __canonical_name__ = "ZMQClientConfig"
    __version__ = SYFT_OBJECT_VERSION_1

    id: Optional[UID]
    hostname: str = "127.0.0.1"


class ZMQClientConfigV2(SyftObject, QueueClientConfig):
    __canonical_name__ = "ZMQClientConfig"
    __version__ = SYFT_OBJECT_VERSION_2

    id: Optional[UID]
    hostname: str = "127.0.0.1"
    queue_port: Optional[int] = None
    # TODO: setting this to false until we can fix the ZMQ
    # port issue causing tests to randomly fail
    create_producer: bool = False
    n_consumers: int = 0


@serializable()
class ZMQClientConfig(SyftObject, QueueClientConfig):
    __canonical_name__ = "ZMQClientConfig"
    __version__ = SYFT_OBJECT_VERSION_3

    id: Optional[UID]
    hostname: str = "127.0.0.1"
    queue_port: Optional[int] = None
    # TODO: setting this to false until we can fix the ZMQ
    # port issue causing tests to randomly fail
    create_producer: bool = False
    n_consumers: int = 0
    consumer_service: Optional[str]


@migrate(ZMQClientConfig, ZMQClientConfigV1)
def downgrade_zmqclientconfig_v2_to_v1():
    return [
        drop(["queue_port", "create_producer", "n_consumers"]),
    ]


@migrate(ZMQClientConfigV1, ZMQClientConfig)
def upgrade_zmqclientconfig_v1_to_v2():
    return [
        make_set_default("queue_port", None),
        make_set_default("create_producer", False),
        make_set_default("n_consumers", 0),
    ]


@serializable(attrs=["host"])
class ZMQClient(QueueClient):
    """ZMQ Client for creating producers and consumers."""

    producers: Dict[str, ZMQProducer]
    consumers: DefaultDict[str, list[ZMQConsumer]]

    def __init__(self, config: ZMQClientConfig) -> None:
        self.host = config.hostname
        self.producers = {}
        self.consumers = defaultdict(list)
        self.config = config

    @staticmethod
    def _get_free_tcp_port(host: str):
        with socketserver.TCPServer((host, 0), None) as s:
            free_port = s.server_address[1]
        return free_port

    def add_producer(
        self,
        queue_name: str,
        port: Optional[int] = None,
        queue_stash=None,
        worker_stash: Optional[WorkerStash] = None,
        context=None,
    ) -> ZMQProducer:
        """Add a producer of a queue.

        A queue can have at most one producer attached to it.
        """

        if port is None:
            if self.config.queue_port is None:
                self.config.queue_port = self._get_free_tcp_port(self.host)
                port = self.config.queue_port
            else:
                port = self.config.queue_port

        producer = ZMQProducer(
            queue_name=queue_name,
            queue_stash=queue_stash,
            port=port,
            context=context,
            worker_stash=worker_stash,
        )
        self.producers[queue_name] = producer
        return producer

    def add_consumer(
        self,
        queue_name: str,
        message_handler: AbstractMessageHandler,
        service_name: str,
        address: Optional[str] = None,
        worker_stash: Optional[WorkerStash] = None,
        syft_worker_id: Optional[UID] = None,
    ) -> ZMQConsumer:
        """Add a consumer to a queue

        A queue should have at least one producer attached to the group.

        """

        if address is None:
            address = get_queue_address(self.config.queue_port)

        consumer = ZMQConsumer(
            queue_name=queue_name,
            message_handler=message_handler,
            address=address,
            service_name=service_name,
            syft_worker_id=syft_worker_id,
            worker_stash=worker_stash,
        )
        self.consumers[queue_name].append(consumer)

        return consumer

    def send_message(
        self,
        message: bytes,
        queue_name: str,
        worker: Optional[bytes] = None,
    ) -> Union[SyftSuccess, SyftError]:
        producer = self.producers.get(queue_name)
        if producer is None:
            return SyftError(
                message=f"No producer attached for queue: {queue_name}. Please add a producer for it."
            )
        try:
            producer.send(message=message, worker=worker)
        except Exception as e:
            # stdlib
            return SyftError(
                message=f"Failed to send message to: {queue_name} with error: {e}"
            )
        return SyftSuccess(
            message=f"Successfully queued message to : {queue_name}",
        )

    def close(self) -> Union[SyftError, SyftSuccess]:
        try:
            for _, consumers in self.consumers.items():
                for consumer in consumers:
                    # make sure look is stopped
                    consumer.close()

            for _, producer in self.producers.items():
                # make sure loop is stopped
                producer.close()
                # close existing connection.
        except Exception as e:
            return SyftError(message=f"Failed to close connection: {e}")

        return SyftSuccess(message="All connections closed.")

    def purge_queue(self, queue_name: str) -> Union[SyftError, SyftSuccess]:
        if queue_name not in self.producers:
            return SyftError(message=f"No producer running for : {queue_name}")

        producer = self.producers[queue_name]

        # close existing connection.
        producer.close()

        # add a new connection
        self.add_producer(queue_name=queue_name, address=producer.address)

        return SyftSuccess(message=f"Queue: {queue_name} successfully purged")

    def purge_all(self) -> Union[SyftError, SyftSuccess]:
        for queue_name in self.producers:
            self.purge_queue(queue_name=queue_name)

        return SyftSuccess(message="Successfully purged all queues.")


@serializable()
class ZMQQueueConfig(QueueConfig):
    def __init__(
        self, client_type=None, client_config=None, thread_workers: bool = False
    ):
        self.client_type = client_type or ZMQClient
        self.client_config: ZMQClientConfig = client_config or ZMQClientConfig()
        self.thread_workers = thread_workers
